# Copyrite IBM 2022, 2025
# IBM Confidential

import hashlib, json, struct, re, base64, binascii, sys, array, os, time, logging, jwt
import cbor2 as cbor
from typing import Optional, List

from cryptography.hazmat.primitives.asymmetric import rsa, ec, ed25519, padding, utils
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.backends import default_backend
from cryptography import x509
from cryptography.fernet import Fernet


try:
    from soft_fido2.key_pair import KeyPair, KeyUtils
    from soft_fido2.cert_utils import CertUtils
    from soft_fido2.symmetric_key import SymmetricKey
except:
    try:
        sys.path.append(os.path.dirname(os.path.abspath(__file__)))
        from cert_utils import CertUtils
        from key_pair import KeyPair, KeyUtils
        from symmetric_key import SymmetricKey    
    except Exception as e:
        logging.debug("Module load error")
        logging.exception(e)
        raise e


class Fido2Authenticator(object):

    def __init__(self,
            keyPair: Optional[KeyPair] = None,
            credId: Optional[bytes] = None,
            aaguid: Optional[List[int]] = None,
            caKeyPair: Optional[KeyPair] = None,
            caCert: Optional[x509.Certificate] = None,
            counter: int = 0,
            hashingAlg: hashes.HashAlgorithm = hashes.SHA256(),
            transports: Optional[List[str]] = None,
            fKey: Optional[Fernet] = None,
            sKey: Optional[SymmetricKey] = None,
            disableCounter: bool = False,
            ctsProfileMatch: bool = True,
            saltLength: Optional[int] = None,
    ) -> None:
        """Initialize a FIDO2 Authenticator with the specified parameters.
        
        Args:
            keyPair (KeyPair, optional): Public/private key pair to sign challenges with.
                    Default = None (EC 256 key will be generated)
            credId (bytes, optional): Credential ID to use with authenticator.
                    Default = None (will be derived from keyPair)
            aaguid (List[int], optional): AAGUID to associate with authenticator.
                    Default = None (will use null aaguid)
            caKeyPair (KeyPair, optional): Public/private key of CA/intermediate authority
                    for BASIC/ATTCA/TPM/Android attestation formats. Default = None
            caCert (x509.Certificate, optional): Certificate to use as a trust anchor.
                    Default = None
            counter (int): Internal counter of token. Default = 0
            hashingAlg (hashes.HashAlgorithm): Hashing algorithm for "packed" attestation.
                    Default = SHA256
            transports (List[str], optional): List of supported transports. Default = None
            fKey (Fernet, optional): Symmetric key to generate credential ID.
                    Can be used to reconstruct private EC key for assertions. Default = None
            sKey (SymmetricKey, optional): Alternative symmetric key to generate credential ID.
                    Can be used to reconstruct private EC key for assertions. Default = None
            disableCounter (bool): Whether to disable the attestation/assertion counter.
                    Default = False
            ctsProfileMatch (bool): Whether to disable the CTS Profile match for Android Safetynet
                                    attestation statements.
            saltLength (int, optional): Length of salt to use for "packed" attestation (RSAPSS).
                    This is expected to be the same length as the digest from the hashing 
                    algorithm used.
        """
        if aaguid and len(aaguid) != 16:
            raise ValueError("AAGUID must be 16 bytes long")
        elif not aaguid:
            aaguid = [0] * 16

        # Core authenticator properties
        self.counter = counter
        self.hashAlg = hashingAlg
        self.disable_counter = disableCounter
        self.aaguid = aaguid
        self.salt_len = saltLength
        self.ctsProfileMatch = ctsProfileMatch

        # Credential and key management
        self.kp = keyPair
        self.fKey = fKey
        self.sKey = sKey
        self.cib = None  # Credential ID bytes

        # Certificate and trust chain
        self.caCertificate = caCert
        self.caKeyPair = caKeyPair
        # Transports
        self.transports = transports

        # Initialize key pair based on credential ID
        self._init_key_pair(credId)

    def _init_key_pair(self, credId):
        if credId is not None:
            self.cib = self._urlb64_decode(credId)
            key = self.fKey or self.sKey
            if self.kp is None and key is not None:
                try:
                    self.kp = self._get_key_pair_from_credential_id(credId, key)
                except Exception:
                    pass
        # Generate a new key pair if we still don't have one
        if self.kp is None:
            self.kp = KeyPair.generate_ecdsa()

    @classmethod
    def _urlb64_decode(cls, b64String):
        """Helper function to decode b64 urlencoded strings which may be missing
        the trailing padding that Python requires.

        Args:
            b64String (str): URL-safe base64 encoded string to decode

        Returns:
            bytes: Decoded bytes
        """
        pad = len(b64String) % 4
        if pad:
            b64String += b'=' * pad
        return base64.urlsafe_b64decode(b64String)

    @classmethod
    def _urlb64_encode(cls, byteString):
        """Helper function or b64 encode a string then remove the trailing padding
        which is not required

        Args:
            byteString (str): string to encode

        Returns:
            str: b64 url encoded string with trailing '=' stripped
        """
        b64String = str(base64.urlsafe_b64encode(byteString), 'utf-8')
        return re.sub(r'=*$', '', b64String)

    @classmethod
    def _long_to_bytes(cls, l):
        """Convert a long to a byte representation

        Args:
            l (long): long to convert to bytes

        Returns:
            :obj:`list` of :obj:`bytes`: byte representation of the long value
        """
        limit = 256**4 - 1  #max value we can fit into a struct.pack
        parts = []
        while l:
            parts.append(l & limit)
            l >>= 32
        parts = parts[::-1]
        return struct.pack(">" + 'L' * len(parts), *parts)

    def _bytes_to_long(self, b):
        """Converts an array of bytes to a long

        Args:
            b (:obj:`list` of :obj:`byte`): bytes to convert

        Returns:
            long: value of bytes as a long
        """
        l = len(b) / 4
        parts = struct.unpack(">" + 'L' * int(l), b)[::-1]
        result = 0
        for i in range(len(parts)):
            temp = parts[i] << (32 * i)
            result += temp

        return result

    def _get_credential_id_bytes(self, keyPair):
        """Get the bytes of a credential ID for a given authenticator.
        If a self.caKeyPair is not None, credId is the encoded bytes of self.kp.get_private

        else credId is the sha256 of the public key

        Args:
            keyPair (KeyPair): key pair to generate Id for
            caKyePair (KeyPair): ca key pair used for encrypting private key, may be None

        Return:
            bytes: credential Id for given key pair and ca key pair
        """
        if self.cib != None:
            return self.cib
        key = (self.sKey or self.fKey) #skey preferred
        if key and isinstance(keyPair.get_public(), ec.EllipticCurvePublicKey):
            #logging.debug(f"Encryping cred id with {key}")
            self.cib = key.encrypt(keyPair.get_private_bytes())
            #logging.debug(f"self.cib {self.cib}")
            return self.cib
        else:
            #logging.debug("using hash of public key as cib")
            return hashlib.sha256(keyPair.get_public_bytes()).digest()

    def get_credential_id(self, keyPair=None):
        """credential ID defaults to the SHA256 of the public key

        Args:
            keyPair (:obj:`KeyPair`, optional): key pair to get credential id for; default = self.kp

        Returns:
            str: b64 encoded byte string of credentail id
        """
        if keyPair == None:
            keyPair = self.kp
        credIdBytes = self._get_credential_id_bytes(keyPair)
        return self._urlb64_encode(credIdBytes)


    @classmethod
    def _get_key_pair_from_credential_id(cls, credId, decryptor):
        """Given a credId and caKeyPair attempt to reconstruct the private/public
        key pair

        Args:
            credId (str): url safe base64 encoded private key encrypted using keyPair
            decryptor (Union[Fernet,SymmetricKey]): key previously used to encrypt private key

        Return:
            KeyPair: original key pair stored in credId
        """
        encBytes = cls._urlb64_decode(credId)
        #decrypt the bytes using the given key
        keyBytes = decryptor.decrypt(encBytes)
        #Finally reconstruct the key
        return KeyPair.load_key_pair(keyBytes)

    def get_aaguid(self, hexString=True):
        """If hexString returns in the format:
            01020304-0506-0708-0900-010203040506

        else returns format:
            1234567890123456

        Args:
            hexString (:obj:`bool`, optional): toggle wether to output a hexstring or a string representation of
                    aaguid; default = True

        Returns:
            str: representation of aaguid
        """
        result = ''
        if hexString:
            for x in range(16):
                result += binascii.hexlify(bytes(chr(self.aaguid[x]), 'utf-8')).decode('utf-8')
                if x == 3 or x == 5 or x == 7 or x == 9:
                    result += '-'
        else:
            result = bytes(self.aaguid)
        return result

    def credential_create(self, jsonOptions, atteStmtFmt='packed-self', keyPair=None, uv=True, up=True, be=False, bs=False):
        '''Responds to requests to navigator.credential.create(). jsonOptions should be
        either a dictionary or a JSON string of the attestation options and usually has the form:
        {
            "rp": {
                "id": "relying.party",
                "name": "Relying Party"
            },
            "user": {
                "id": "my_unique_id",
                "name": "Low Key",
                "displayName": "redacted"
            },
            "timeout": 60000,
            "challenge": "wvhbvWMV5Jsl96WbdZGav6Ifpp8QHnJC0MKhs1vDUes",
            "excludeCredentials": [],
            "authenticatorSelection": {
                "requireResidentKey": true,
                "authenticatorAttachment": "cross-platform",
                "userVerification": "preferred"
            },
            "attestation": "direct",
            "pubKeyCredParams": [
                {
                    "alg": -7,
                    "type": "public-key"
                },
                {
                    "alg": -257,
                    "type": "public-key"
                }
            ]
        }

        Args:
            jsonOptions (dict or str): Dictionary or JSON string of options for navigator.credential.create
            atteStmtFmt (str, optional): Attestation statement format as defined in
                    https://w3c.github.io/webauthn/#defined-attestation-formats
                    Defaults to 'packed-self'.
                    For compound attestation, the string is prefixed with 'compound:' and the required compound
                    statements are listed with a ',' (comma) separator. e.g.: `compound:packed-self,tpm`
            keyPair (KeyPair, optional): Private/public key pair to sign the attestation. Defaults to self.kp.
            uv (bool, optional): Whether the authenticator should set the user verification flag. Defaults to True.
            up (bool, optional): Whether the authenticator should set the user presence flag. Defaults to True.
            be (bool, optional): Whether the authenticator should set the backup eligible flag. Defaults to False.
            bs (bool, optional): Whether the authenticator should set the backup state flag. Defaults to False.

        Returns:
            dict: Response to navigator.credential.create containing the attestation
        '''
        if keyPair is None:
            keyPair = self.kp
        options = {}
        if isinstance(jsonOptions, dict):
            options = jsonOptions
        else:
            options = json.loads(jsonOptions)
        cco = self.attestation_options_response_to_credential_create_options(options)
        return self.process_credential_create_options(cco, atteStmtFmt, keyPair, uv, up)

    def credential_request(self, jsonOptions, keyPair=None, uv=True, up=True, be=False, bs=False):
        '''Responds to navigator.credential.get(). jsonOptions should be either a dictionary
        or a JSON string of the assertion options and usually has the form:
        {
            "rpID": "www.my-relying-party.com"
            "userId": "my_unique_id",
            "displayName": "redacted",
            "authenticatorSelection": {
                "requireResidentKey": false,
                "authenticatorAttachment": "cross-platform",
                "userVerification": "preferred"
            },
            "attestation": "direct"
        }

        If you want to use a different origin you can add it to the Public-Key dictionary as a 
        top level key.

        Args:
            jsonOptions (dict): json dictionary of options for navigator.credentials.get
            keyPair (:obj:`KeyPair`, optional): private/public key pair to sign the assertion; default = self.kp
            uv (:obj:`bool`, optional): if the authenticator should set hte user verification flag, default = True
            up (:obj:`bool`, optional): if the authenticator should set hte user presence flag, default = True
            be (:obj:`bool`, optional): if the authenticator should set the backup eligible flag; default = False
            bs (:obj:`bool`, optional): if the authenticator should set the backup state flag; default = False

        Returns:
            dict: response to navigator.credential.get
        '''
        if keyPair is None:
            keyPair = self.kp
        options = {}
        if isinstance(jsonOptions, dict):
            options = jsonOptions
        else:
            options = json.loads(jsonOptions)
        cro = self.assertion_options_response_to_credential_request_options(options)

        return self.process_credential_request_options(cro, keyPair, uv, up, be, bs)

    def build_client_data_JSON(self, pk):
        """Creates the ClientDataJSON object for attestation and assertion operations

        Args:
            pk (dict): public key dictionary from request options,
                    https://www.w3.org/TR/webauthn/#dictdef-publickeycredentialcreationoptions
                    https://www.w3.org/TR/webauthn/#dictdef-publickeycredentialrequestoptions

        Returns:
            dict: clientDataJSON, https://www.w3.org/TR/webauthn/#sec-client-data
        """
        mode = 'webauthn.get' if 'rpId' in pk else 'webauthn.create'
        origin = pk.get('origin', None)
        if not origin and 'rpId' in pk:
            origin = 'https://' + pk['rpId']
        if not origin:
            origin = 'https://' + pk['rp']['id']

        clientDataDict = {'origin': origin, 'challenge': self._urlb64_encode(pk['challenge']), 'type': mode}
        return json.dumps(clientDataDict)

    def process_attested_credential_data(self, publicKey, credIdBytes):
        """create the attested credentail data for attestation requets

        Args:
            publickey: (PublicKey): RSA || EC public key
            credIdBytes (str): byte string of credential id, https://www.w3.org/TR/webauthn/#credential-id

        Returns:
            str: attested credetail data, https://www.w3.org/TR/webauthn/#sec-attested-credential-data
        """
        attestedCredDataBytes = []
        attestedCredDataBytes += array.array('B', self.aaguid).tobytes()
        length = struct.pack('H', len(credIdBytes))
        attestedCredDataBytes += [length[1], length[0]]
        attestedCredDataBytes += credIdBytes
        credPublicKeyCOSE = KeyUtils.get_cose_key(publicKey, self.hashAlg)
        attestedCredDataBytes += cbor.dumps(credPublicKeyCOSE)
        return attestedCredDataBytes

    def build_authenticator_data(self, pk, attStmtFmt, keyPair, uv, up=True, be=False, bs=False):
        """Create the authenticator data for the attestation or assertion request.

        Args:
            pk (dict): Public key dictionary from request options.
                    For attestation: https://www.w3.org/TR/webauthn/#dictdef-publickeycredentialcreationoptions
                    For assertion: https://www.w3.org/TR/webauthn/#dictdef-publickeycredentialrequestoptions
            attStmtFmt (str): Attestation statement format as defined in
                    https://www.w3.org/TR/webauthn/#defined-attestation-formats
            keyPair (KeyPair): Public/private key pair to use
            uv (bool): Whether to set the user verification flag
            up (bool, optional): Whether to set the user presence flag. Defaults to True.
            be (bool, optional): Whether to set the backup eligible flag. Defaults to False.
            bs (bool, optional): Whether to set the backup state flag. Defaults to False.

        Returns:
            bytes: Authenticator data as defined in
                  https://www.w3.org/TR/webauthn/#sec-authenticator-data
        """
        authDataBytes = []

        rpId = pk.get('rpId', None)
        assertion = True
        if not rpId:
            rpId = pk['rp']['id']
            assertion = False

        rpIdHash = hashlib.sha256(rpId.encode('utf-8')).digest()
        authDataBytes += rpIdHash

        flags = 0x00
        if up:
            flags |= 0x01  # UP
        if not assertion:
            flags |= 0x40  # AT
        if attStmtFmt != 'fido-u2f' and uv != None and uv == True:
            flags |= 0x04  # UV
        if be == True:
            flags |= 0x08
        if bs == True:
            flags |= 0x0F
        authDataBytes += struct.pack("c", chr(flags).encode('utf-8'))

        #Add counter and increment if required
        authDataBytes += struct.pack(">I", self.counter)
        if self.disable_counter == False:  
            self.counter += 1

        if not assertion:
            credIdBytes = self._get_credential_id_bytes(keyPair)
            authDataBytes += self.process_attested_credential_data(keyPair.get_public(), credIdBytes)
        authData = bytes(authDataBytes)
        return authData

    def build_packed_attestation_statement(self, atteStmtFmt, clientDataHash, authData, credIdBytes, keyPair):
        """Create an attestation statment with the packed format. Hashing alg used in this format is controleld by
        the `self.hashAlg` property

        Args:
            atteStmtFmt (str): statement format, either 'packed' or 'packed-self' to indicate self signed attestation
            clientDataHash (str): byte string of clientDataHash,
                    https://www.w3.org/TR/webauthn/#collectedclientdata-hash-of-the-serialized-client-data
            authData (str): byte string of the authentication data,
                    https://www.w3.org/TR/webauthn/#sec-authenticator-data
            credIdBytes (str): byte string of the credential id
            keyPair (KeyPair): public/privte key pair to sign data with

        Returns:
            dict: packed attestation statement,
                    https://www.w3.org/TR/webauthn/#packed-attestation
        """
        result = {} # Key order is important
        result[u"alg"] = KeyUtils.get_alg_id_from_pubkey_and_hash(keyPair.get_public(), self.hashAlg, pss=True if self.salt_len else False)
        toSign = bytes([*authData, *clientDataHash])
        sig = ""

        if isinstance(keyPair.get_public(), rsa.RSAPublicKey):
            #sig = keyPair.get_private().sign( toSign, padding.PKCS1v15(), hashes.SHA256() )
            sig = keyPair.get_private().sign(toSign, padding.PKCS1v15(), self.hashAlg)

        elif isinstance(keyPair.get_public(), ec.EllipticCurvePublicKey):
            #digest = hashes.Hash(hashes.SHA256())
            digest = hashes.Hash(self.hashAlg)
            digest.update(b''.join([(x.encode() if isinstance(x, str) else bytes([x])) for x in toSign]))
            #sig = keyPair.get_private().sign( digest.finalize(), ec.ECDSA(utils.Prehashed(hashes.SHA256())) )
            sig = keyPair.get_private().sign(digest.finalize(), ec.ECDSA(utils.Prehashed(self.hashAlg)))
        elif isinstance(keyPair.get_public(), ed25519.Ed25519PublicKey):
            sig = keyPair.get_private().sign(toSign)
        else:
            # TODO promote PQC to 1st class
            try:
                from oqs.oqs import Signature
                if isinstance(keyPair.get_private(), Signature):
                    sig = keyPair.get_private().sign(toSign)
                else:
                    raise RuntimeError("Unsupported key type")
            except ImportError:
                raise Exception("Unsupported key type")

        result[u"sig"] = sig

        #Maybe add X5c
        selfAttestation = True if 'self' in atteStmtFmt else False
        if not selfAttestation:
            if not self.caCertificate:
                raise RuntimeError("Packed Attestation requires a CA certificate to be "\
                        "present when the authenticator is created")
            leafSubj = x509.Name([
                x509.NameAttribute(x509.NameOID.COMMON_NAME, u'leaf'),
                x509.NameAttribute(x509.NameOID.ORGANIZATIONAL_UNIT_NAME, u'Authenticator Attestation'),
                x509.NameAttribute(x509.NameOID.COUNTRY_NAME, u'AU'),
                x509.NameAttribute(x509.NameOID.ORGANIZATION_NAME, u'IBM')
            ])
            leafCert = CertUtils.gen_aik_cert(subject=leafSubj,
                                              issuer=self.caCertificate.subject,
                                              keyPair=keyPair,
                                              signKeyPair=self.caKeyPair,
                                              aaguid=self.get_aaguid(hexString=False))
            # Final trust chain to add to AttesationObject
            result['x5c'] = [CertUtils.get_encoded(leafCert), CertUtils.get_encoded(self.caCertificate)]
        return result

    def build_fido_u2f_attestation_statement(self, atteStmtFmt, clientDataHash, authData, credIdBytes, keyPair):
        """Create an attestation statement with the U2F format.

        Args:
            atteStmtFmt (str): statement format
            clientDataHash (str): byte string of clientDataHash,
                    https://www.w3.org/TR/webauthn/#collectedclientdata-hash-of-the-serialized-client-data
            authData (str): byte string of the authentication data,
                    https://www.w3.org/TR/webauthn/#sec-authenticator-data
            credIdBytes (str): byte string of the credential id
            keyPair (KeyPair): public/privte key pair to sign data with

        Returns:
            dict: u2f attestation statement,
                    https://www.w3.org/TR/webauthn/#fido-u2f-attestation

        """
        if not isinstance(keyPair.get_public(), ec.EllipticCurvePublicKey):
            raise Exception("FIDO U2F only supports ECDSA keys")

        pubKey = bytearray(b'\x04')
        pubKey.extend(self._long_to_bytes(keyPair.get_public().public_numbers().x))
        pubKey.extend(self._long_to_bytes(keyPair.get_public().public_numbers().y))

        subject = x509.Name([
            x509.NameAttribute(x509.NameOID.COMMON_NAME, u'root'),
            x509.NameAttribute(x509.NameOID.ORGANIZATIONAL_UNIT_NAME, u'IBM Security')
        ])
        cert = CertUtils.gen_ca_cert(subject=subject, keyPair=keyPair)

        rpIdHash = authData[0:32]
        toSign = []
        toSign += ['\x00']
        toSign += rpIdHash
        toSign += clientDataHash
        toSign += credIdBytes
        toSign += pubKey
        digest = hashes.Hash(hashes.SHA256())
        digest.update(b''.join([(x.encode() if isinstance(x, str) else bytes([x])) for x in toSign]))
        sig = keyPair.get_private().sign(digest.finalize(), ec.ECDSA(utils.Prehashed(hashes.SHA256())))
        result = {'sig': sig, 'x5c': [CertUtils.get_encoded(cert)]}

        return result

    def _build_rsa_public_area(self, keyPair):
        pubArea = []
        pubArea += [0, 1]  # TPM_ALG_ID = TPM_ALG_RSA
        pubArea += [0, 11]  # name_alg = TPM_ALG_SHA256
        pubArea += [0] * 4  # TPMA_OBJECT
        pubArea += [0] * 2  # authPolicy
        pubArea += [0, 0x10]  # symetric = TPM_ALG_NULL
        pubArea += [1, 4]  # scheme = TMP_ALG_RSASSA (PKCS1-v1.5)
        pubArea += [4, 0]  # keySize
        pubArea += [0] * 4  # exponent
        unique = self._long_to_bytes(keyPair.get_public().public_numbers().n)
        uniqueLength = struct.pack("!H", len(unique))
        pubArea += [uniqueLength[0], uniqueLength[1]]
        pubArea += unique
        return bytes(pubArea)

    def _build_ec_public_area(self, keypair):
        pubArea = []
        pubArea += [0, 0x23] # TPM_ALG_ID = TPM_ALG_ECC
        pubArea += [0, 0x0B] # TPM_ALG_SHA256
        pubArea += [0] * 4  # TPMA_OBJECT
        pubArea += [0] * 2  # authPolicy
        pubArea += [0, 0x10]  # symetric = TPM_ALG_NULL
        pubArea += [0, 0x10]  # scheme = TPM_ALG_NULL
        pubArea += [0, 0x03]  # curve_id == TPM_ECC_NIST_P256
        pubArea += [0, 0x10]  # kdf == TPM_ALG_NULL
        xBytes = KeyUtils._long_to_bytes(keypair.get_public().public_numbers().x)
        xByteLen = struct.pack("!H", len(xBytes))
        pubArea += [xByteLen[0], xByteLen[1]]
        pubArea += xBytes
        yBytes = KeyUtils._long_to_bytes(keypair.get_public().public_numbers().y)
        yByteLen = struct.pack("!H", len(yBytes))
        pubArea += [yByteLen[0], yByteLen[1]]
        pubArea += yBytes
        return bytes(pubArea)

    def _build_cert_info(self, attsToSign, pubInfo):
        certInfo = [0xFF, 0x54, 0x43, 0x47]  # TPM_GENERATED
        certInfo += [0x80, 0x17]  # TPM_ST_ATTEST_CERTIFY
        certInfo += [0] * 2  # qualified signer length
        digest = hashes.Hash(hashes.SHA256())
        digest.update(attsToSign)
        sigHash = digest.finalize()
        #certInfo += [ int((len(sigHash) - (len(sigHash) & 0xFF)) / 256), len(sigHash) & 0xFF ]
        sigHashLength = struct.pack("!H", len(sigHash))
        certInfo += [sigHashLength[0], sigHashLength[1]]
        certInfo += sigHash
        certInfo += [0] * 17  # clock info
        vendorId = struct.pack("!L", CertUtils.TPM_VENDOR_ID)
        certInfo += [0] * (8 - len(vendorId))
        certInfo += vendorId
        attestedName = [0x00, 0x0B]  #name_alg
        digest = hashes.Hash(hashes.SHA256())
        digest.update(pubInfo)
        attestedName += digest.finalize()
        attestedNameLength = struct.pack("!H", len(attestedName))
        certInfo += [attestedNameLength[0], attestedNameLength[1]]
        certInfo += attestedName
        certInfo += [0] * 2  # attested qualified name length
        return bytes(certInfo)

    def build_tpm_attestation_statement(self, atteStmtFmt, clientDataHash, authData, credIdBytes, keyPair):
        """Create an attestation statement with the TPM format.

        Args:
            atteStmtFmt (str): Statement format ('tpm')
            clientDataHash (bytes): Hash of the client data as defined in
                    https://www.w3.org/TR/webauthn/#collectedclientdata-hash-of-the-serialized-client-data
            authData (bytes): Authenticator data as defined in
                    https://www.w3.org/TR/webauthn/#sec-authenticator-data
            credIdBytes (bytes): Credential ID bytes
            keyPair (KeyPair): Public/private key pair to sign data with

        Returns:
            dict: TPM attestation statement as defined in
                  https://www.w3.org/TR/webauthn/#tpm-attestation
        """
        if not self.caCertificate:
            raise RuntimeError("TPM Attestation requires a CA certificate to be "\
                    "present when the authenticator is created")
        #Generate TPM certificates
        vendorId = CertUtils._long_to_bytes(CertUtils.TPM_VENDOR_ID).hex()
        tpmSan = x509.Name([
            x509.NameAttribute(x509.ObjectIdentifier(CertUtils.TPM_MANUFACTURER), u"id:{}".format(vendorId)),
            x509.NameAttribute(x509.ObjectIdentifier(CertUtils.TPM_VENDOR), u"IBMTPM"),
            x509.NameAttribute(x509.ObjectIdentifier(CertUtils.TPM_FW_VERSION), u"id:1")
        ])
        tpmCert = CertUtils.gen_aik_cert(subject=x509.Name([]),
                                         issuer=self.caCertificate.subject,
                                         keyPair=keyPair,
                                         signKeyPair=self.caKeyPair,
                                         aaguid=self.get_aaguid(hexString=False),
                                         san=tpmSan)
        x5c = [CertUtils.get_encoded(tpmCert), CertUtils.get_encoded(self.caCertificate)]

        # Build sign data
        toSign = bytes([*authData, *clientDataHash])
        pubArea = self._build_rsa_public_area(keyPair) if isinstance(keyPair.get_public(), rsa.RSAPublicKey) else \
                    self._build_ec_public_area(keyPair)
        certInfo = self._build_cert_info(toSign, pubArea)
        sig = None
        theHash = hashes.SHA256()
        if isinstance(keyPair.get_public(), rsa.RSAPublicKey):
            keyPair.get_private().sign(certInfo, padding.PKCS1v15(), theHash)
            sig = keyPair.get_private().sign(certInfo, padding.PKCS1v15(), theHash)
        else:
            digest = hashes.Hash(hashes.SHA256())
            digest.update(certInfo)
            sig = keyPair.get_private().sign(digest.finalize(),
                                                  ec.ECDSA(utils.Prehashed(theHash)) )


        # Build attestation
        result = {
            u"pubArea": pubArea,
            u"certInfo": certInfo,
            u"sig": sig,
            u"ver": u"2.0",
            u"alg": KeyUtils.get_alg_id_from_pubkey_and_hash(keyPair.get_public(), theHash),
            u"x5c": x5c
        }
        return result

    def build_none_attestation_statement(self, atteStmtFmt, clientDataHash, authData, credIdBytes, keyPair):
        """Create an attestation statement with the none format

        Args:
            atteStmtFmt (str): statement format
            clientDataHash (str): byte string of clientDataHash,
                    https://www.w3.org/TR/webauthn/#collectedclientdata-hash-of-the-serialized-client-data
            authData (str): byte string of the authentication data,
                    https://www.w3.org/TR/webauthn/#sec-authenticator-data
            credIdBytes (str): byte string of the credential id
            keyPair (KeyPair): public/privte key pair to sign data with

        Returns:
            dict: none attestation statement,
                    https://www.w3.org/TR/webauthn/#none-attestation
        """
        return {}

    def build_android_safetynet_attestation_statement(self, atteStmtFmt, clientDataHash, authData, credIdBytes,
                                                      keyPair):
        """Create an attestation statement with the Android Safetynet format

        Args:
            atteStmtFmt (str): statement format
            clientDataHash (str): byte string of clientDataHash,
                    https://www.w3.org/TR/webauthn/#collectedclientdata-hash-of-the-serialized-client-data
            authData (str): byte string of the authentication data,
                    https://www.w3.org/TR/webauthn/#sec-authenticator-data
            credIdBytes (str): byte string of the credential id
            keyPair (KeyPair): public/privte key pair to sign data with

        Returns:
            dict: Android safetynet attestation statement,
                    https://www.w3.org/TR/webauthn/#android-safetynet-attestation
        """
        if(isinstance(keyPair.get_public(), ec.EllipticCurvePublicKey)):
           raise RuntimeError("Android safetynet Attestation requires a RSA key")
        if self.caCertificate is None:
            raise RuntimeError("Android safetynet Attestation requires a CA certificate to be" + \
                    " set for the authenticator")
        leafSubj = x509.Name([
            x509.NameAttribute(x509.NameOID.COMMON_NAME, u'attest.android.com'),
            x509.NameAttribute(x509.NameOID.ORGANIZATIONAL_UNIT_NAME, u'Authenticator Attestation'),
            x509.NameAttribute(x509.NameOID.COUNTRY_NAME, u'AU'),
            x509.NameAttribute(x509.NameOID.ORGANIZATION_NAME, u'IBM')
        ])
        leafCert = CertUtils.gen_aik_cert(subject=leafSubj,
                                          issuer=self.caCertificate.subject,
                                          keyPair=keyPair,
                                          signKeyPair=self.caKeyPair)
        nonceBytes = [*authData, *clientDataHash]
        nonceHash = hashlib.sha256(bytes(nonceBytes)).digest()
        claims = {
            u'timestampMs': round(time.time() * 1000),
            u'nonce': base64.b64encode(nonceHash).decode(),
            u'apkPackageName': u"com.package.name.of.requesting.app",
            u"apkCertificateDigestSha256": [u"b64 encoded sha256 of cert"],
            u"ctsProfileMatch": self.ctsProfileMatch,
            u"basicIntegrity": True
        }
        jwtResponse = jwt.encode(
            claims,
            keyPair.get_private_bytes(),
            algorithm="RS256",
            headers={"x5c": [CertUtils.get_bytes(leafCert).decode(),
                             CertUtils.get_bytes(self.caCertificate).decode()]})
        result = {u'ver': u'some version', u'response': jwtResponse.encode()}
        return result

    def build_android_key_attestation_statement(self, atteStmtFmt, clientDataHash, authData, credIdBytes, keyPair):
        """Create an attestation statement with the Android Keystore format.

        Args:
            atteStmtFmt (str): statement format
            clientDataHash (str): byte string of clientDataHash,
                    https://www.w3.org/TR/webauthn/#collectedclientdata-hash-of-the-serialized-client-data
            authData (str): byte string of the authentication data,
                    https://www.w3.org/TR/webauthn/#sec-authenticator-data
            credIdBytes (str): byte string of the credential id
            keyPair (KeyPair): public/privte key pair to sign data with

        Returns:
            dict: Android Keystore attestation statement,
                    https://www.w3.org/TR/webauthn/#android-key-attestation
        """
        if not self.caCertificate:
            raise RuntimeError("Android Key Attestation requires a CA certificate to be "\
                    "present when the authenticator is created")

        #Build x5c chain
        leafSubj = x509.Name([
            x509.NameAttribute(x509.NameOID.COMMON_NAME, u'leaf'),
            x509.NameAttribute(x509.NameOID.ORGANIZATIONAL_UNIT_NAME, u'Authenticator Attestation'),
            x509.NameAttribute(x509.NameOID.COUNTRY_NAME, u'AU'),
            x509.NameAttribute(x509.NameOID.ORGANIZATION_NAME, u'IBM')
        ])
        leafCert = CertUtils.gen_aik_cert(subject=leafSubj,
                                          issuer=self.caCertificate.subject,
                                          keyPair=keyPair,
                                          signKeyPair=self.caKeyPair,
                                          androidKeyNonce=bytes(clientDataHash))
        x5c = [CertUtils.get_encoded(leafCert), CertUtils.get_encoded(self.caCertificate)]
        #Sign data
        toSign = [*authData, *clientDataHash]
        sig = None
        if isinstance(keyPair.get_public(), rsa.RSAPublicKey):
            sig = keyPair.get_private().sign(bytes(toSign), padding.PKCS1v15(), hashes.SHA256())
        else: #Must be EC key
            digest = hashes.Hash(hashes.SHA256())
            digest.update(bytes(toSign))
            sig = keyPair.get_private().sign(digest.finalize(),
                                                  ec.ECDSA(utils.Prehashed(hashes.SHA256())) )

        result = {
            u"x5c": x5c,
            u"sig": sig,
            u"alg": KeyUtils.get_alg_id_from_pubkey_and_hash(keyPair.get_public(), self.hashAlg)
        }
        return result

    def build_apple_attestation_statement(self, atteStmtFmt, clientDataHash, authData, credIdBytes, keyPair):
        """Create an attestation statement with the Apple Platform format.

        Args:
            atteStmtFmt (str): Statement format ('apple')
            clientDataHash (bytes): Hash of the client data as defined in
                    https://www.w3.org/TR/webauthn/#collectedclientdata-hash-of-the-serialized-client-data
            authData (bytes): Authenticator data as defined in
                    https://www.w3.org/TR/webauthn/#sec-authenticator-data
            credIdBytes (bytes): Credential ID bytes
            keyPair (KeyPair): Public/private key pair to sign data with

        Returns:
            dict: Apple platform attestation statement containing x5c certificate chain
        """
        if not self.caCertificate:
            raise RuntimeError("Apple Attestation requires a CA certificate to be "\
                    "present when the authenticator is created")
        #First need to generate the apple certificate with the required extension
        nonceBytes = []
        nonceBytes += authData
        nonceBytes += clientDataHash
        nonceHash = hashlib.sha256(bytes(nonceBytes)).digest()
        leafSubj = x509.Name([
            x509.NameAttribute(x509.NameOID.COMMON_NAME, u'apple'),
            x509.NameAttribute(x509.NameOID.ORGANIZATIONAL_UNIT_NAME, u'Authenticator Attestation'),
            x509.NameAttribute(x509.NameOID.COUNTRY_NAME, u'AU'),
            x509.NameAttribute(x509.NameOID.ORGANIZATION_NAME, u'IBM')
        ])
        appleCert = CertUtils.gen_apple_cert(subject=leafSubj,
                                             issuer=self.caCertificate.subject,
                                             keyPair=keyPair,
                                             signKeyPair=self.caKeyPair,
                                             nonce=nonceHash)
        print(CertUtils.get_encoded(appleCert))
        return {'x5c': [CertUtils.get_encoded(appleCert), CertUtils.get_encoded(self.caCertificate)]}

    def _error(self, atteStmtFmt, clientDataHash, authData, credIdBytes, keyPair):
            raise ValueError("Unsupported attestation statement format")

    def _process_att_stmt(self, atteStmtFmt, clientDataHash, authData, credIdBytes, keyPair):
        return {
            "none": self.build_none_attestation_statement,
            "packed": self.build_packed_attestation_statement,
            "fido-u2f": self.build_fido_u2f_attestation_statement,
            "packed-self": self.build_packed_attestation_statement,
            "android-key": self.build_android_key_attestation_statement,
            "android-safetynet": self.build_android_safetynet_attestation_statement,
            "tpm": self.build_tpm_attestation_statement,
            "apple": self.build_apple_attestation_statement
        }.get(atteStmtFmt, self._error)(atteStmtFmt, clientDataHash, authData, credIdBytes, keyPair)

    def process_attestation_statement(self, atteStmtFmt, clientDataHash, authData, credIdBytes, keyPair):
        """Helper function that chooses an attestation statement function based on the atteStmtFmt variable

        Args:
            atteStmtFmt (str): statement format
            clientDataHash (str): byte string of clientDataHash,
                    https://www.w3.org/TR/webauthn/#collectedclientdata-hash-of-the-serialized-client-data
            authData (str): byte string of the authentication data,
                    https://www.w3.org/TR/webauthn/#sec-authenticator-data
            credIdBytes (str): byte string of the credential id
            keyPair (KeyPair): public/privte key pair to sign data with

        Returns:
            dict: attestation statement. Type of statement depends on 'atteStmtFmt', see:
                    https://www.w3.org/TR/webauthn/#defined-attestation-formats
        """
        if atteStmtFmt.startswith('compound'):
            stmtsCsv = atteStmtFmt.split(":")
            if stmtsCsv == None or len(stmtsCsv) != 2:
                raise Exception("Unexpected attestation statement format [{}]".format(atteStmtFmt))
            stmts = stmtsCsv[1].split(",")
            if stmts == None or len(stmts) < 2:
                raise Exception("Unexpected attestation statement format [{}]".format(atteStmtFmt))
            result = []
            for stmt in stmts:
                result += [{u'fmt': stmt, u'attStmt': 
                           self._process_att_stmt(stmt, clientDataHash, authData, credIdBytes, keyPair)}]
            return result
        #else
        return self._process_att_stmt(atteStmtFmt, clientDataHash, authData, credIdBytes, keyPair)

    def attestation_options_response_to_credential_create_options(self, options):
        """Take the options provided by the relying party and extract required information to
        generate the attestation.

        Args:
            options (dict): Options from navigator.credential.create as defined in
                    https://www.w3.org/TR/webauthn/#credentialcreationoptions-extension

        Returns:
            dict: Credential creation options as defined in
                  https://www.w3.org/TR/webauthn/#dictionary-makecredentialoptions
        """
        pkcco = {'rp': options['rp']}
        user = {'id': self._urlb64_decode(options['user']['id'].encode('UTF-8'))}
        pkcco['user'] = user
        pkcco['challenge'] = self._urlb64_decode(options['challenge'].encode('UTF-8'))
        pkcco['pubKeyCredParams'] = options['pubKeyCredParams']
        if 'timeout' in options:
            pkcco['timeout'] = options['timeout']

        if 'excludeCredentials' in options:
            pkcco['excludeCredentials'] = options['excludeCredentials']

        if 'authenticatorSelection' in options:
            pkcco['authenticatorSelection'] = options['authenticatorSelection']

        if 'attestation' in options:
            pkcco['attestation'] = options['attestation']

        if 'extensions' in options:
            pkcco['extensions'] = options['extensions']

        if 'origin' in options:
            pkcco['origin'] = options['origin']

        cco = {'publicKey': pkcco}
        return cco

    def process_credential_create_options(self, cco, atteStmtFmt, keyPair, uv, up=True, be=False, bs=False):
        """Generate response to parsed credential create request

        Args:
            cco (dict): Credential Create Options,
                    https://www.w3.org/TR/credential-management-1/#credentialcreationoptions-dictionary
            atteStmtFmt (str): required attestation format. see:
                    https://www.w3.org/TR/webauthn/#defined-attestation-formats
            keyPair (KeyPair): public/private kye pair to sign with
            uv (bool): set the user verification flag
            up (bool): set the user presence flag
            be (bool): set the backup eligible flag
            bs (bool): set the backup state flag

        Returns:
            dict: attestation response to credential create request,
                    https://www.w3.org/TR/webauthn/#authenticatorattestationresponse
        """
        pk = cco['publicKey']
        self.userHandle = pk['user']['id']
        clientDataJSON = self.build_client_data_JSON(pk)
        clientDataHash = hashlib.sha256(clientDataJSON.encode('utf-8')).digest()
        clientDataEncoded = base64.urlsafe_b64encode(clientDataJSON.encode('ascii'))

        credIdBytes = self._get_credential_id_bytes(keyPair)

        authData = self.build_authenticator_data(pk, atteStmtFmt, keyPair, uv, up, be, bs)
        attStmt = self.process_attestation_statement(atteStmtFmt, clientDataHash, authData, credIdBytes, keyPair)
        attStmtFmt = str(re.sub('-self', '', atteStmtFmt))
        if atteStmtFmt.startswith('compound'):
            attStmtFmt = atteStmtFmt.split(':')[0]
        attestationObject = {u'authData': authData, u'fmt': attStmtFmt, u'attStmt': attStmt}
        saar = {
            u'clientDataJSON': str(clientDataEncoded, 'utf-8'),
            u'attestationObject': str(base64.urlsafe_b64encode(cbor.dumps(attestationObject)), 'utf-8')
        }
        spkc = {
            u'id': self.get_credential_id(keyPair),
            u'rawId': self.get_credential_id(keyPair),
            u'response': saar,
            u'type': u'public-key',
            u'getClientExtensionResults': {}
        }
        if self.transports is not None:
            spkc['getTransports'] = self.transports
        if(cco.get('extensions') != None 
                and isinstance(cco['extensions'], dict) 
                and "devicePubKey" in cco['extensions'].keys()):
            raise RuntimeError("TODO")
        return spkc

    def assertion_signature(self, authData, clientDataHash, keyPair):
        """Generate a signature for an assertion using the appropriate algorithm for the key type.

        Args:
            authData (bytes): Authenticator data
            clientDataHash (bytes): Hash of the client data
            keyPair (KeyPair): Public/private key pair to sign with

        Returns:
            bytes: Signature of the combined authenticator data and client data hash

        Raises:
            Exception: If the key algorithm is not supported
        """
        toSign = []
        toSign += authData
        toSign += clientDataHash
        toSignStr = bytes(toSign)
        sig = b''
        if isinstance(keyPair.get_public(), rsa.RSAPublicKey) == True:
            if self.salt_len:
                sig = keyPair.get_private().sign(toSignStr, padding.PSS(mgf=padding.MGF1(self.hashAlg), salt_length=self.salt_len), self.hashAlg)
            else:
                sig = keyPair.get_private().sign(toSignStr, padding.PKCS1v15(), self.hashAlg)
        elif isinstance(keyPair.get_public(), ec.EllipticCurvePublicKey) == True:
            hasher = hashes.Hash(self.hashAlg)
            hasher.update(toSignStr)
            sig = keyPair.get_private().sign(hasher.finalize(),
                                                  ec.ECDSA(utils.Prehashed(self.hashAlg)))
        elif isinstance(keyPair.get_public(), ed25519.Ed25519PublicKey):
            sig = keyPair.get_private().sign(toSignStr)
        else:
            raise Exception("Unsupported key algorithm")
        return sig

    def assertion_options_response_to_credential_request_options(self, options):
        """Take the options provided by the relyig party and extract required information to
        generate the assertion

        Args:
            options (dict): options from navigator.credential.get
                    https://www.w3.org/TR/webauthn/#iface-authenticatorassertionresponse
        Returns:
            dict: https://www.w3.org/TR/credential-management-1/#dictdef-credentialrequestoptions
        """
        cro = {}
        pkcro = {}

        pkcro['challenge'] = self._urlb64_decode(options['challenge'].encode('UTF-8'))
        if 'timeout' in options:
            pkcro['timeout'] = options['timeout']

        pkcro['rpId'] = options['rpId']
        if 'allowedCredentials' in options:
            allowedCreds = options['allowedCredentials']
            pkcro['allowedCredentials'] = []
            for c in allowedCreds:
                cred = {'type': c['type'], 'id': base64.urlsafe_b64decode(c['id'])}
                if 'transports' in c:
                    cred['transports'] = c['transports']
                pkcro['allowedCredentials'].append(cred)

        if 'userVerifation' in options:
            pkcro['userVerification'] = options['userVerification']

        if 'extensions' in options:
            pkcro['extensions'] = options['extensions']

        if 'origin' in options:
            pkcro['origin'] = options['origin']

        cro['publicKey'] = pkcro
        return cro

    def process_credential_request_options(self, cro, keyPair, uv, up=True, be=False, bs=False):
        """Generate response to parsed credential get request

        Args:
            cro (dict): Credential Request Options,
                    https://www.w3.org/TR/credential-management-1/#dictdef-credentialrequestoptions
            keyPair (KeyPair): public/private key pair to sign with
            uv (bool): set the user verification flag
            up (bool): set the user presence flag
            be (bool): set the backup eligible flag. This should be consistent with the registration state.
            bs (bool): set the backup state flag

        Returns:
            dict: assertion response to credential get request,
                    https://www.w3.org/TR/webauthn/#authenticatorassertionresponse
        """
        pk = cro["publicKey"]
        clientDataJSON = self.build_client_data_JSON(pk)
        authData = self.build_authenticator_data(pk, None, keyPair, uv, up, be, bs)
        saar = {
            "clientDataJSON": str(base64.urlsafe_b64encode(clientDataJSON.encode('utf-8')), 'utf-8'),
            "authenticatorData": str(base64.urlsafe_b64encode(authData), 'utf-8')
        }
        if hasattr(self, 'userHandle') and isinstance(self.userHandle, (str, bytes)):
            saar['userHandle'] = self._urlb64_encode(self.userHandle)
        if "attestation" in cro.keys():
            raise RuntimeError("TODO")
        clientDataHash = bytearray(hashlib.sha256(clientDataJSON.encode('utf-8')).digest())

        saar['signature'] = str(base64.urlsafe_b64encode(self.assertion_signature(
                                                            authData, clientDataHash, keyPair)), 'utf-8')
        spkc = {
            'id': self.get_credential_id(keyPair),
            'rawId': self.get_credential_id(keyPair),
            'response': saar,
            'type': 'public-key',
            'getClientExtensionResults': {}
        }
        if(cro.get('extensions', None) != None
                and isinstance(cro['extensions'], dict)
                and "devicePubKey" in cro['extensions'].keys()):
            raise RuntimeError("TODO")
        return spkc


############################# MAIN ##############################

if __name__ == "__main__":
    if "FIDO_HOME" not in os.environ:
        logging.debug("Cannot find passkey home \"FIDO_HOME\"")
        sys.exit(1)
    authenticator = Fido2Authenticator()
    rsp = None
    pubPath = os.path.join(os.environ['FIDO_HOME'], 'test_public.pem')
    pivPath = os.path.join(os.environ['FIDO_HOME'], 'test_private.pem')
    if sys.argv[1] == 'attestation':
        if authenticator.kp is not None:
            rsp = authenticator.credential_create(sys.argv[3], 
                                                                  atteStmtFmt=sys.argv[2], 
                                                                  keyPair=authenticator.kp)
            with open(pivPath, 'wb') as key_file:
                key_file.write(authenticator.kp.get_private_bytes())
            with open(pubPath, 'wb') as key_file:
                key_file.write(authenticator.kp.get_public_bytes())
    elif sys.argv[1] == 'assertion':
        privateKey = publicKey = None
        with open(pivPath, 'rb') as key_file:
            privateKey = serialization.load_pem_private_key(key_file.read(), 
                                                                             password=None,
                                                                             backend=default_backend())

        with open(pubPath, 'rb') as key_file:
            publicKey = serialization.load_pem_public_key(key_file.read(),
                                                                          backend=default_backend())

        keyPair = KeyPair(privateKey, publicKey)
        authenticator.kp = keyPair
        rsp = authenticator.credential_request(sys.argv[2], authenticator.kp)
    else:
        logging.debug("Must specify a ceeremony (attestation || assertion) to perform.")
        sys.exit(1)
    print(json.dumps(rsp, indent=4))



#!/bin/python
#IBM Confidential
# Assisted by watsonx Code Assistant
# Copyright IBM Corp. 2025

import os, struct, fcntl, time, queue, threading, logging, re

from enum import Enum

try:
    from soft_fido2.message_queues import QueueMessageType, MessageQueue
except:
    from message_queues import QueueMessageType, MessageQueue

# Assisted by watsonx Code Assistant 
#logging.basicConfig(filename='passkey.log', filemode='a', level=logging.DEBUG, format='%(asctime)s - %(levelname)s - %(message)s')

# Thanks StackOverflow !
class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    OKPINK = '\033[95m'
    OKYELLOW = '\033[93m'
    OKPURPLE = '\033[35m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def colour_print(colour=bcolors.OKBLUE, component='USB/IP', msg=''):
    logging.debug('[' + colour + component + bcolors.ENDC + '] ' + msg)


def print_bytes(*args):
    result = ""
    count = 0
    for ba in args:
        for x in ba:
            result += "%02X " % x
            count += 1
            if count == 8 :
                result += " "
            elif count == 16:
                logging.debug("\t" + result)
                result = ""
                count = 0
    logging.debug('\t' + result + '\n')

def dump_bytes(*args, colour=bcolors.OKPURPLE, component='UHID Device', msg=''):
    #Print bytes in nice format
    c = colour if colour != None else bcolors.OKPURPLE
    colour_print(colour=colour, component=component, msg=msg)
    print_bytes(*args)


UHID_EVENT_TYPE_SIZE = 4
'''
uint_32_t type;
'''

EV_MAX_SIZE = 4380
'''
/**
 * Fedora 40 test.
 * $ gcc -o ./uhid_test -Wall -I./include ./uhid-test.c
 *
 * $ ./uhid_test
 * 4380
 */
// uhid-test.c
#include <linux/uhid.h>
#include <unistd.h>
#include <stdio.h>

int main(int argc, char **argv)
{
    struct uhid_event ev;
    size_t ev_size = sizeof(ev);
    fprintf(stderr, "%ld\n", ev_size);
    return 0;
}
'''


class UHIDEventType(Enum):
    CREATE = 0x00
    DESTROY = 0x01
    START = 0x02
    STOP = 0x03
    OPEN = 0x04
    CLOSE = 0x05
    OUTPUT = 0x06
    GET_REPORT = 0x09
    GET_REPORT_REPLY = 0x0A
    CREATE2 = 0x0B
    INPUT2 = 0x0C
    SET_REPORT = 0x0D
    SET_REPORT_REPLY = 0x0E


    @classmethod
    def from_bytes(cls, byte_data):
        if len(byte_data) != 4:
            raise ValueError("Expected 4 bytes to parse UHIDEventType")
        event_int = struct.unpack('I', byte_data)[0]
        try:
            return cls(event_int)
        except ValueError:
            raise ValueError(f"Unknown UHIDEventType value: {event_int}")

    def pack(self):
        return struct.pack('I', self.value)


class UHIDReportType(Enum):
    FEATURE_REPORT = 0x00
    OUTPUT_REPORT = 0x01
    INPUT_REPORT = 0x02

class BaseStructure(object):
    _fields_ = []
    base_pack_format = '<'

    def __init__(self, **kwargs):
        self.init_from_dict(**kwargs)
        for field in self._fields_:
            if len(field) > 2:
                if not hasattr(self, field[0]):
                    setattr(self, field[0], field[2])

    def init_from_dict(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def size(self):
        return struct.calcsize(self.format())

    def format(self):
        pack_format = self.base_pack_format
        for field in self._fields_:
            if isinstance(field[1], BaseStructure):
                pack_format += str(field[1].size()) + 's'
            elif 'si' == field[1]:
                pack_format += 'c'
            elif '<' in field[1] or '>' in field[1]:
                pack_format += field[1][1:]
            else:
                pack_format += field[1]
        logging.debug(pack_format)
        return pack_format.encode('utf-8')

    def pack(self):
        values = []
        for field in self._fields_:
            #logging.debug("Field: {}".format(field))
            if isinstance(field[1], BaseStructure):
                values.append(getattr(self, field[0], field[1]).pack())
            elif re.match(r'\d*x', field[1]):
                #Skip padding
                continue
            else:
                if 'si' == field[1]:
                    values.append(chr(getattr(self, field[0], 0)))
                else:
                    values.append(getattr(self, field[0], 0))
        #Python 2 -> 3, str != bytestring so conditionally remap any strings we find.
        values = [bytes(v, 'utf-8') if isinstance(v, str) else v for v in values]
        logging.debug(values)
        packed = struct.pack(self.format(), *values)
        #logging.debug("packed [{}]".format(packed))
        return packed

    def unpack(self, buf):
        values = struct.unpack(self.format(), buf)
        i=0
        keys_vals = {}
        for val in values:
            if '<' in self._fields_[i][1][0]:
                val = struct.unpack('<' +self._fields_[i][1][1], struct.pack('>' + self._fields_[i][1][1], val))[0]
            keys_vals[self._fields_[i][0]]=val
            i+=1
        #logging.debug(keys_vals)
        self.init_from_dict(**keys_vals)


REPORT_DESCRIPTOR = bytes([
    0x06, 0xD0, 0xF1, # Usage Page (FIDO Alliance)
    0x09, 0x01, # Usage (U2F HID Authenticator)
    0xA1, 0x01, # Collection (Application)

    0x09, 0x20, # Usage (Input Report Data)
    0x15, 0x00, # Logical Minimum (0)
    0x26, 0xFF, 0x00, # Logical Maximum (255)
    0x75, 0x08,             # REPORT_SIZE (8)
    0x95, 0x40,             # REPORT_COUNT (64)
    0x81, 0x02, # Input (Data,Var,Abs)

    0x09, 0x21, # Usage (Output Report Data)
    0x15, 0x00, # Logical Minimum (0)
    0x26, 0xFF, 0x00, # Logical Maximum (255)
    0x75, 0x08,             # REPORT_SIZE (8)
    0x95, 0x40,             # REPORT_COUNT (64)
    0x91, 0x02, # Output (Data,Var,Abs)

    0xC0 # End Collection
])

DEVICE_NAME = b"EyeBeeKey"
PHYSICAL_ADDRESS = b"ibm-0101:01:01.0-1"
UNIQUE_ADDRESS = b"virtual-fido-uhid-01"

class UHIDCreate2Event(BaseStructure):
    _fields_ = [
            ('event', 'I', UHIDEventType.CREATE2.value),
            ('name', '128s', DEVICE_NAME.ljust(128, b'\x00')),
            ('phys', '64s', PHYSICAL_ADDRESS.ljust(64, b'\x00')),
            ('uniq', '64s', UNIQUE_ADDRESS.ljust(64, b'\x00')),
            ('rd_size', 'H', len(REPORT_DESCRIPTOR)),
            ('bus', 'H', 0x03), # USB HID
            ('vendor', 'I', 0x1337),
            ('product', 'I', 0x1337),
            ('version', 'I', 0x0100), # Version 1.00
            ('country', 'I', 0), # Not localized
            ('rd_data', '4096s', REPORT_DESCRIPTOR.ljust(4096, b'\x00'))
        ]

class UHIDStartEvent(BaseStructure):
    _fields_ = [
            ('event', 'I', UHIDEventType.START.value),
            ('dev_flags', 'Q')
        ]

class UHIDStopEvent(BaseStructure):
    _fields_ = [
            ('event', 'I', UHIDEventType.STOP.value)
        ]

class UHIDInput2Event(BaseStructure):

    _fields_ = [
            ('event', 'I', UHIDEventType.INPUT2.value),
            ('ev_len', 'H', 0),
            ('data', '4096s', b'\x00' * 4096 )
        ]

class UHIDOutputEvent(BaseStructure):
    ev_len = 0
    type = UHIDReportType.OUTPUT_REPORT.value
    data = b'\x00'

    _fields_ = [
            ('event', 'I', UHIDEventType.OUTPUT.value),
            ('data', '4096s', b'\x00' * 4096),
            ('ev_len', 'H', 0),
            ('type', 'B', UHIDReportType.OUTPUT_REPORT.value)
        ]


class UHIDOpenEvent(BaseStructure):
    _fields_ = [
            ('event', 'I', UHIDEventType.OPEN.value)
        ]

class UHIDCloseEvent(BaseStructure):
    _fields_ = [
            ('event', 'I', UHIDEventType.CLOSE.value)
        ]

class UHIDGetReport(BaseStructure):
    _fields_ = [
            ('event', 'I', UHIDEventType.GET_REPORT.value),
            ('id', 'I'),
            ('report_number', 'B'),
            ('report_type', 'B')
        ]

class UHIDGetReportReply(BaseStructure):
    _fields_ = [
            ('event', 'I', UHIDEventType.GET_REPORT_REPLY.value),
            ('id', 'I'),
            ('err', 'H'),
            ('ev_len', 'H'),
            ('data', '4096s')
        ]

class UHIDSetReport(BaseStructure):
    report_type = 0x01

    _fields_ = [
            ('event', 'I', UHIDEventType.SET_REPORT.value),
            ('id', 'I'),
            ('report_number', 'B'),
            ('report_type', 'B'),
            ('ev_len', 'H'),
            ('data', '4096s')
        ]

class UHIDSetReportReply(BaseStructure):
    _fields_ = [
            ('event', 'I', UHIDEventType.SET_REPORT_REPLY.value),
            ('id', 'I'),
            ('err', 'H')
        ]

class UserDevice(threading.Thread):


    #Pending input reports
    pending = queue.Queue(maxsize=100)
    events = []

    def __init__(self, devPath="/dev/uhid"):
        super().__init__()
        self.device_path = devPath
        self._interrupt = False
        #import signal
        #signal.signal(signal.SIGINT, self.stop_device)
        #signal.signal(signal.SIGTERM, self.stop_device)



    def stop_device(self, tid, frame):
        logging.error(f"Received interrupt from {tid}: frame {frame}")
        self._interrupt = True
        if MessageQueue.notify_sysapp is not None:
            MessageQueue.notify_sysapp.put(QueueMessageType.QUIT)

    # Assisted by watsonx Code Assistant 
    def format_bytes(self, byte_array):
        return ''.join(f'{byte:02x}' for byte in byte_array)

    # Assisted by watsonx Code Assistant 
    def log_received_bytes(self, byte_array, io_type="IN"):
        logging.debug("event byte count: {} DIR: {}".format( 
                            0 if byte_array is None else len(byte_array), io_type))
        formatted_bytes = self.format_bytes(byte_array)
        lines = [formatted_bytes[i:i+64] for i in range(0, len(formatted_bytes), 64)]
        for line in lines:
            logging.debug(line)

    def start_ev(self, ev_type, ev_bytes):
        logging.debug("Start event received!")
        ev = UHIDStartEvent()
        ev.unpack(ev_bytes[:len(ev.pack())])
        logging.debug(f"ev: {ev}")
        return

    def stop_ev(self, ev_type, ev_bytes):
        logging.debug("Stop event received")
        ev = UHIDStopEvent()
        ev.unpack(ev_bytes[:len(ev.pack())])
        logging.debug(f"ev: {ev}")
        return

    def open_ev(self, ev_type, ev_bytes):
        logging.debug("Open event received!")
        ev = UHIDOpenEvent()
        ev.unpack(ev_bytes[:len(ev.pack())])
        logging.debug(f"ev: {ev}")
        return

    def close_ev(self, ev_type, ev_bytes):
        logging.debug("Close event received!")
        ev = UHIDCloseEvent()
        ev.unpack(ev_bytes[:len(ev.pack())])
        logging.debug(f"ev: {ev}")
        MessageQueue.notify_auth.put(QueueMessageType.CLOSE_EVENT)
        return

    def process_output(self, event):
        raise NotImplementedError("override me")

    def output_ev(self, ev_type, ev_bytes):
        logging.debug("Output event received!")
        ev = UHIDOutputEvent()
        ev.unpack(ev_bytes[:len(ev.pack())])
        if ev.data:
            logging.debug(f"event : {ev.ev_len} {ev.type} {ev.data[:ev.ev_len]}")
        self.process_output(ev)

    def handle_unknown_control(self, ev_bytes):
        raise NotImplementedError("override me")

    def get_report_ev(self, ev_type, ev_bytes):
        logging.debug("Get report event received!")
        ev = UHIDGetReport()
        ev.unpack(ev_bytes[:len(ev.pack())])
        logging.debug(f"ev: {ev}")
        raise NotImplementedError("unexpected request")

    def set_report_ev(self, ev_type, ev_bytes):
        logging.debug("Set report event received!")
        ev = UHIDSetReport()
        ev.unpack(ev_bytes[:len(ev.pack())])
        logging.debug(f"ev: {ev}")
        if ev.report_type == UHIDReportType.FEATURE_REPORT:
            raise NotImplementedError("TODO")
        else:
            raise NotImplementedError("unexpected request")

    def error_ev(self, ev_type, ev_bytes):
        logging.debug("Unknown UHID event received")
        logging.debug(f"ev_bytes: {ev_bytes}")
        return

    def process_event(self, ev_type, ev_bytes):
        return {
                UHIDEventType.START: self.start_ev,
                UHIDEventType.STOP: self.stop_ev,
                UHIDEventType.OPEN: self.open_ev,
                UHIDEventType.CLOSE: self.close_ev,
                UHIDEventType.OUTPUT: self.output_ev,
                UHIDEventType.GET_REPORT: self.get_report_ev,
                UHIDEventType.SET_REPORT: self.set_report_ev
            }.get(ev_type, self.error_ev)(ev_type, ev_bytes)

    def destroy_ev(self, fd):
        ev = UHIDInput2Event(event=UHIDEventType.DESTROY.value, ev_len=0)
        os.write(fd, ev.pack())

    def _maybe_in(self, fd):
        try: #Poll for event
            eventBytes = os.read(fd, EV_MAX_SIZE)
            #self.log_received_bytes(eventBytes)
            if isinstance(eventBytes, bytes) \
                    and len(eventBytes) >= UHID_EVENT_TYPE_SIZE:
                eventType = UHIDEventType.from_bytes(
                                eventBytes[:UHID_EVENT_TYPE_SIZE])
                thread = threading.Thread(target=self.process_event, args=(eventType, eventBytes))
                thread.start()
                self.events.append(thread)
            else:
                logging.error(f"Invalid event read [{eventBytes}]")
        except BlockingIOError: #No event
            #logging.debug("No data available (non-blocking read)") #very very very verbose
            pass
        tempEvList = []
        for ev in self.events:
            if not ev.is_alive():
                ev.join()
                tempEvList.append(ev)
        for ev in tempEvList:
            self.events.remove(ev)

    def _maybe_out(self, fd):
        try:
            while self.pending.qsize() > 0: # uhid device sends output to kernel
                inData = self.pending.get(True, 0.00001) #10ns
                ev = UHIDInput2Event(ev_len=64, data=inData)
                inBytes = ev.pack()
                #self.log_received_bytes(inBytes, io_type="OUT")
                n = os.write(fd, bytearray(inBytes))
                if n != len(inBytes):
                    raise RuntimeError(f"invalid write length {n} != {len(ev.pack())}")
                else:
                    logging.debug("Event sent!")
                logging.debug(f"{self.pending.qsize()} events left in the queue")
        except queue.Empty:
            logging.debug("Failed to get more output events, not sending anything else.")


    def run(self):
        fd = None
        started = False
        try:
            fd = os.open('/dev/uhid', os.O_RDWR)  #| os.O_CLOEXEC| os.O_NONBLOCK
            fcntl.fcntl(fd, fcntl.F_SETFL, os.O_NONBLOCK)
        except OSError as e:
            logging.exception(f"OSError with uhid fd: {e}")
            return
        if fd == None:
            logging.error("udev fd is null")
            return
        try:
            #Send create
            create_2_req = UHIDCreate2Event().pack()
            n = os.write(fd, bytearray(create_2_req))
            if n != len(create_2_req):
                raise RuntimeError("invalid write length")
            while not self._interrupt:
                self._maybe_in(fd)
                self._maybe_out(fd)
                if not self._interrupt:
                    time.sleep(0.001) #poll respective queues every ms
                if MessageQueue.notify_udev.qsize() > 0:
                    sysTrayMsg = MessageQueue.notify_udev.get()
                    logging.debug(f"Event from systeray_app: {sysTrayMsg}") 
                    if sysTrayMsg == QueueMessageType.QUIT:
                        self._interrupt = True
                        break
        finally:
            if started:
                self.destroy_ev(fd)
            os.close(fd)



# Copyrite IBM 2022, 2025
# IBM Confidential

import os, time, sys, subprocess, traceback, shutil, threading, logging, signal
from PyQt6.QtGui import QIcon, QAction
from PyQt6.QtWidgets import (QApplication, QDialog, QDialogButtonBox, QHBoxLayout,
                QLabel, QLineEdit, QListWidget, QMessageBox, QPushButton, QSystemTrayIcon, 
                QMenu, QVBoxLayout, QComboBox)
from PyQt6.QtCore import (QObject, QRunnable, QThreadPool, pyqtSignal, pyqtSlot, 
                        QTimer, Qt, QMetaObject)
try:
    from soft_fido2.message_queues import QueueMessageType, MessageQueue
    from soft_fido2.key_pair import KeyUtils
except:
    from message_queues import QueueMessageType, MessageQueue
    from key_pair import KeyUtils

class WorkerSignals(QObject):
    # Define signals as class attributes here
    error = pyqtSignal(tuple)

class Worker(QRunnable):
    def __init__(self, handle, *args, **kwargs):
        super().__init__()
        self.handle = handle
        self.args = args
        self.kwargs = kwargs
        self.signals = WorkerSignals()

    @pyqtSlot()
    def run(self):
        try:
            self.handle(*self.args, **self.kwargs)
        except Exception:
            traceback.print_exc()
            exctype, value = sys.exc_info()[:2]
            self.signals.error.emit((exctype, value, traceback.format_exc()))


class ManageCredentialsDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Manage Credentials")
        
        # Initialize instance variables
        self.fido_home = os.environ.get('FIDO_HOME', os.path.expanduser('~/.fido'))
        self.credentials = []
        
        # Create main layout
        layout = QVBoxLayout()
        
        # Add UI components to layout
        layout.addLayout(self._create_pin_input_section())
        layout.addLayout(self._create_passkey_selection_section())
        layout.addWidget(QLabel("Credentials:"))
        layout.addWidget(self._create_credentials_list())
        layout.addLayout(self._create_action_buttons())
        layout.addWidget(self._create_close_button())
        
        self.setLayout(layout)

    def _get_passkey_files(self):
        """
        Returns a list of .passkey files in the specified directory.
        
        Args:
            directory: The directory to search for .passkey files
            
        Returns:
            A list of full paths to .passkey files
        """
        passkey_files = []
        for filename in os.listdir(self.fido_home):
            if filename.endswith('.passkey'):
                passkey_files.append(os.path.join(self.fido_home, filename))
        return passkey_files

    def _create_pin_input_section(self):
        """Create the PIN input section with label and password field."""
        layout = QHBoxLayout()
        layout.addWidget(QLabel("PIN:"))
        self.pin_input = QLineEdit()
        self.pin_input.setEchoMode(QLineEdit.EchoMode.Password)
        layout.addWidget(self.pin_input)
        return layout
    
    def _create_passkey_selection_section(self):
        """Create the passkey selection section with label and dropdown."""
        layout = QHBoxLayout()
        layout.addWidget(QLabel("Passkey:"))
        self.name_input = QComboBox()
        
        # Populate dropdown with passkey files
        self._load_passkey_files()
        
        layout.addWidget(self.name_input)
        return layout
    
    def _create_credentials_list(self):
        """Create the credentials list widget."""
        self.creds_list = QListWidget()
        return self.creds_list
    
    def _create_action_buttons(self):
        """Create the Load and Delete buttons."""
        layout = QHBoxLayout()
        
        self.load_button = QPushButton("(re)Load Credentials (and Cache pin)")
        self.load_button.clicked.connect(self.load_credentials)
        
        self.delete_button = QPushButton("Delete Resident Credential")
        self.delete_button.clicked.connect(self.delete_credential)
        
        layout.addWidget(self.load_button)
        layout.addWidget(self.delete_button)
        return layout
    
    def _create_close_button(self):
        """Create the Close button."""
        close_button = QPushButton("Close")
        close_button.clicked.connect(self.reject)
        return close_button
    
    def _load_passkey_files(self):
        """Load passkey files and populate the dropdown."""
        if os.path.exists(self.fido_home):
            passkey_files = self._get_passkey_files()
            for passkey_file in passkey_files:
                # Add just the basename to the dropdown
                self.name_input.addItem(os.path.basename(passkey_file))

    def _try_cache_pin(self, nonce, passkey_path):
        """ Load then Save passkey. This will update the cached upper pin hash
        with the provided secret if it successfully unpacks the passkey file. """
        self.passkey = KeyUtils._load_passkey(nonce, passkey_path)
        self.credentials = self.passkey['res.creds']
        KeyUtils._save_passkey(
            self.passkey['key'],
            self.passkey['x5c'],
            self.passkey['res.creds'],
            self.passkey['pin.hash'],
            passkey_path
        )

    def load_credentials(self):
        try:
            # Get PIN hash
            pin = self.pin_input.text()
            nonce = KeyUtils.get_pin_hash(pin)
            # Get the selected passkey filename
            passkey_name = self.name_input.currentText()
            passkey_path = os.path.join(self.fido_home, passkey_name)
            
            self._try_cache_pin(nonce, passkey_path)
            # Success
            # Clear the list widget
            self.creds_list.clear()
            
            # Add credentials to the list widget showing only rp.id and user.id
            for cred in self.credentials:
                # Convert rp.id and user.id to something printable
                rp_id_value = cred.get('rp.id', 'cred.parsing.error')
                user_id_value = cred.get('user.id', 'cred.parsing.error')
                
                # RpId should be utf-8 compatible; user-id is probably not...
                rp_id = rp_id_value.decode('utf-8') if isinstance(rp_id_value, bytes) else str(rp_id_value)
                user_id = user_id_value.hex() if isinstance(user_id_value, bytes) else str(user_id_value)
                # Add to list widget
                item_text = f"uri: {rp_id} | user id: {user_id}"
                self.creds_list.addItem(item_text)
                
            
        except Exception as e:
            logging.exception(f"failed to load the credentials from {self.name_input.currentText()} : {e}")
            self.creds_list.clear()
            self.passkey = None
            self.credentials = []
            self.pin_input.clear()
            self.pin_input.setFocus()
        
    def delete_credential(self):
        # Get selected items
        selected_items = self.creds_list.selectedItems()
        if not selected_items:
            return
        
        # Confirm deletion
        confirm = QMessageBox.question(
            self,
            "Confirm Deletion",
            f"Are you sure you want to delete {len(selected_items)} selected credential(s)?",
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
            QMessageBox.StandardButton.No
        )
        
        if confirm != QMessageBox.StandardButton.Yes:
            return
        
        # Get indices of selected items
        selected_indices = [self.creds_list.row(item) for item in selected_items]
        
        # Remove credentials from the list (in reverse order to avoid index shifting)
        for index in sorted(selected_indices, reverse=True):
            if 0 <= index < len(self.credentials):
                del self.credentials[index]
        self.write_passkey()

    def write_passkey(self): 
        if not self.passkey:
            return
        # Update the passkey and save it back to disk
        self.passkey['res.creds'] = self.credentials
        try:
            # Use the same passkey path that was used for loading
            passkey_path = os.path.join(self.fido_home, self.name_input.currentText())
            
            KeyUtils._save_passkey(
                self.passkey['key'],
                self.passkey['x5c'],
                self.credentials,
                self.passkey['pin.hash'],
                passkey_path
            )
            
            # Reload the credentials list to reflect changes
            self.load_credentials()
            
            QMessageBox.information(
                self,
                "Success",
                "Selected credentials have been deleted successfully."
            )
        except Exception as e:
            logging.exception(f"Failed to save updated passkey: {e}")
            QMessageBox.critical(
                self,
                "Error",
                f"Failed to delete credentials: {str(e)}"
            )

class CollectPlatformSecretDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Collect Platform Secret")

        layout = QVBoxLayout()
        
        # Pin
        pin_layout = QHBoxLayout()
        pin_layout.addWidget(QLabel("Passphrase: "))
        self.pin_input = QLineEdit()
        self.pin_input.setEchoMode(QLineEdit.EchoMode.Password)
        pin_layout.addWidget(self.pin_input)
        layout.addLayout(pin_layout)
        
        # Filename
        filename_layout = QHBoxLayout()
        filename_layout.addWidget(QLabel("Filename: "))
        self.filename_input = QLineEdit()
        self.filename_input.setText("platform.key")
        filename_layout.addWidget(self.filename_input)
        layout.addLayout(filename_layout)

        # Buttons
        button_box = QDialogButtonBox(
                QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        button_box.accepted.connect(self.accept)
        button_box.rejected.connect(self.reject)
        layout.addWidget(button_box)

        self.setLayout(layout)
    
    def get_pin(self):
        return self.pin_input.text()
    
    def get_filename(self):
        return self.filename_input.text()

    def reject(self):
        super().reject()

class GeneratePasskeyDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Generate New Passkey")
        layout = QVBoxLayout()
        
        # PIN input with password masking
        pin_layout = QHBoxLayout()
        pin_layout.addWidget(QLabel("PIN:"))
        self.pin_input = QLineEdit()
        self.pin_input.setEchoMode(QLineEdit.EchoMode.Password)
        self.pin_input.setPlaceholderText("Default: 00000000")
        pin_layout.addWidget(self.pin_input)
        layout.addLayout(pin_layout)
        
        # Passkey filename input
        name_layout = QHBoxLayout()
        name_layout.addWidget(QLabel("Passkey name:"))
        self.name_input = QLineEdit()
        self.name_input.setPlaceholderText("Default: default")
        name_layout.addWidget(self.name_input)
        layout.addLayout(name_layout)
        
        # Buttons
        button_box = QDialogButtonBox(
                QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        button_box.accepted.connect(self.accept)
        button_box.rejected.connect(self.reject)
        layout.addWidget(button_box)
        
        self.setLayout(layout)
    
    def get_values(self):
        pin = self.pin_input.text() or "00000000"
        passkey_name = self.name_input.text() or "default"
        return pin, passkey_name

class SysTrayApp(QDialog):
    class NotificationFramework:
        NOTIFY_SEND = 0
        QT = 1
    
    # Global flag for signal handling
    _received_signal = False
    _signal_num = 0
    
    def __init__(self):
        self.app = QApplication(sys.argv)
        super().__init__()
        
        self.main_icon = self._generate_icon('../icons/main_icon.png',
                                            QIcon.ThemeIcon.DialogPassword)
        self.auth_icon = self._generate_icon('../icons/auth_request.png',
                                            QIcon.ThemeIcon.DialogWarning)
        
        # Create the tray icon as a member variable
        self._tray_icon = QSystemTrayIcon(self.main_icon, self)
        self._tray_icon.setToolTip('EyeBeeKey')
        
        self.menu = self._menu_setup()
        self.notification_fw = self._setup_notifications()
        self.threadPool = self._threadpool_setup()
        self.worker = self._worker_setup()
        self.quit = False
        
        # Track active dialog to prevent multiple dialogs
        self._active_dialog = None
        
        # Create a timer to reset the icon after a period of time
        self.icon_reset_timer = QTimer(self)
        self.icon_reset_timer.setSingleShot(True)
        self.icon_reset_timer.timeout.connect(self._reset_icon)
        
        # Set up signal handling with Qt
        self._setup_signal_handling()
        
        # Hide the dialog window by default
        self.hide()
        
        self._finalise()
    
    def _setup_signal_handling(self):
        """Set up signal handling"""
        # Set up the signal handlers
        signal.signal(signal.SIGINT, SysTrayApp._signal_handler)
        signal.signal(signal.SIGTERM, SysTrayApp._signal_handler)
        
        # Create a timer to check for signals
        self._signal_timer = QTimer(self)
        self._signal_timer.timeout.connect(self._check_signal)
        self._signal_timer.start(100)  # Check every 100ms
    
    @staticmethod
    def _signal_handler(sig, frame):
        """Signal handler that sets the global flag"""
        logging.info(f"Received signal {sig}, setting flag for Qt event loop")
        try:
            SysTrayApp._received_signal = True
            SysTrayApp._signal_num = sig
        except Exception as e:
            logging.error(f"Error in signal handler: {e}")
    
    def _check_signal(self):
        """Check if a signal has been received"""
        if SysTrayApp._received_signal:
            sig_name = "SIGINT" if SysTrayApp._signal_num == signal.SIGINT else "SIGTERM"
            logging.info(f"Qt event loop detected {sig_name}, shutting down gracefully")
            SysTrayApp._received_signal = False
            self._exit()

    def _setup_notifications(self):
        if shutil.which('notify-send'):
            return self.NotificationFramework.NOTIFY_SEND
        else:
            self._tray_icon.messageClicked.connect(self.on_message_clicked)
            return self.NotificationFramework.QT

    def launch_notification(self):
        return {
            self.NotificationFramework.NOTIFY_SEND: NotifySend.launch_notification,
            self.NotificationFramework.QT: self._launch_notification_fallback
            }.get(self.notification_fw, self._launch_notification_fallback)()

    def prompt_notification(self):
        return {
            self.NotificationFramework.NOTIFY_SEND: NotifySend.prompt_notification,
            self.NotificationFramework.QT: self._prompt_notification_fallback
            }.get(self.notification_fw, self._prompt_notification_fallback)()

    def cancel_notification(self):
        NotifySend.cancel_notification()

    def _launch_notification_fallback(self):
        self._tray_icon.showMessage("EyeBeeKey",
                         "Use the Generate Cache Key option to get started",
                         QSystemTrayIcon.MessageIcon.Information, 3000)

    def _prompt_notification_fallback(self):
        self._tray_icon.showMessage("EyeBeeKey",
                         "Pirate key recieved a webauthn ceremony, should I respond?",
                         QSystemTrayIcon.MessageIcon.Critical, 15000)

    def on_message_clicked(self):
        MessageQueue.notify_auth.put(QueueMessageType.USER_RESPONSE_ACCEPT)

    def _menu_setup(self):
        menu = QMenu()
        action_setup = [
                        self.__generate_platform_key_action_setup,
                        self.__generate_passkey_action_setup,
                        self.__manage_credentials_action_setup,
                        self.__exit_action_setup]
        for action in action_setup:
            menu.addAction(action())
        return menu

    def _generate_icon(self, path, fallback):
        icon = None
        icon_path = os.path.join(os.getcwd(), path)
        if os.path.exists(icon_path):
            icon = QIcon(icon_path)
        else:
            icon = QIcon.fromTheme(fallback)
        return icon

    def __generate_platform_key_action_setup(self):
        action = QAction('Generate Cache Key', self.app)
        action.triggered.connect(self.__generate_platform_key)
        return action

    def __generate_passkey_action_setup(self):
        action = QAction('Generate Passkey', self.app)
        action.triggered.connect(self.__generate_passkey)
        return action

    def __manage_credentials_action_setup(self):
        action = QAction('Manage Credentials', self.app)
        action.triggered.connect(self.__manage_credentials)
        return action

    def __exit_action_setup(self):
        action = QAction('Exit', self.app)
        action.triggered.connect(self._exit)
        return action

    def __generate_platform_key(self):
        '''
        Generate a new platform key in the specified file with the optional given secret.
        '''
        # Check if another dialog is already active
        if self._active_dialog is not None:
            QMessageBox.information(
                self,
                "Operation in Progress",
                "Please complete the current operation before starting a new one."
            )
            return
            
        dialog = CollectPlatformSecretDialog(self)
        # Connect to the dialog's signals
        dialog.accepted.connect(lambda: self.__handle_platform_key_dialog(dialog))
        dialog.rejected.connect(lambda: self.__handle_dialog_closed(dialog))
        
        # Set as active dialog
        self._active_dialog = dialog
        dialog.show()
            
    def __validate_filename_input(self, filename):
        # Ensure filename has .key extension
        if not filename.endswith('.key'):
            filename += '.key'
        
        fido_home = os.environ.get('FIDO_HOME', os.path.expanduser('~/.fido'))
        os.makedirs(fido_home, exist_ok=True)
        return os.path.join(fido_home, filename)


    def __handle_platform_key_dialog(self, dialog):
        '''
        Handle the platform key dialog's accepted signal.
        '''
        # Clear active dialog reference
        self._active_dialog = None
        try:
            pin = dialog.get_pin()
            filename = dialog.get_filename()
            
            platform_key_path = self.__validate_filename_input(filename)
            # Check if file already exists            
            if os.path.exists(platform_key_path):
                # Ask for confirmation before overwriting
                confirm = QMessageBox.question(
                    self,
                    "Confirm Overwrite",
                    f"File {filename} already exists. Overwrite?",
                    QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                    QMessageBox.StandardButton.No
                )
                if confirm != QMessageBox.StandardButton.Yes:
                    return
            
            # Create the platform key
            nonce = pin if pin and len(pin) > 0 else None
            KeyUtils.create_platform_key(secret=nonce, filename=filename)
            
            QMessageBox.information(
                self,
                "Success",
                f"Platform key created successfully as {filename}"
            )
            
        except Exception as e:
            logging.error(f"Error processing platform key: {e}")
            QMessageBox.critical(
                self,
                "Error",
                f"Failed to create platform key: {str(e)}"
            )

    def __generate_passkey(self):
        # Check if another dialog is already active
        if self._active_dialog is not None:
            QMessageBox.information(
                self,
                "Operation in Progress",
                "Please complete the current operation before starting a new one."
            )
            return
            
        dialog = GeneratePasskeyDialog(self)
        # Connect to the dialog's signals
        dialog.accepted.connect(lambda: self.__handle_generate_passkey_dialog(dialog))
        dialog.rejected.connect(lambda: self.__handle_dialog_closed(dialog))
        
        # Set as active dialog
        self._active_dialog = dialog
        dialog.show()
        
    def __handle_generate_passkey_dialog(self, dialog):
        # Clear active dialog reference
        self._active_dialog = None
        pin, passkey_name = dialog.get_values()
        try:
            passkey_data = KeyUtils.generate_passkey()
            pin_hash = KeyUtils.get_pin_hash(pin)
            # Save passkey
            fido_home = os.environ.get("FIDO_HOME", None)
            if not fido_home:
                self._tray_icon.showMessage("EyeBeeKey",
                     "FIDO_HOME property not set, restart passkey process",
                     QSystemTrayIcon.MessageIcon.Critical, 7500)
                return
            os.makedirs(fido_home, exist_ok=True)
            passkey_path = os.path.join(fido_home, f"{passkey_name}.passkey")
            KeyUtils._save_passkey(
                passkey_data['key'],
                passkey_data['x5c'],
                [],  # No resident credentials initially
                pin_hash,
                passkey_path
            )
            QMessageBox.information(None,
                                    "Success",
                                    f"Passkey {passkey_name}.passkey created in {fido_home}")
    
        except Exception as e:
            QMessageBox.critical(None,
                                 "Error",
                                 f"Failed to generate passkey: {str(e)}")
            self._exit()

    def __manage_credentials(self):
        # Check if another dialog is already active
        if self._active_dialog is not None:
            QMessageBox.information(
                self,
                "Operation in Progress",
                "Please complete the current operation before starting a new one."
            )
            return
            
        dialog = ManageCredentialsDialog(self)
        # No special handling needed for this dialog's result
        # but we should still clean up when it's closed
        dialog.finished.connect(lambda: self.__handle_dialog_closed(dialog))
        
        # Set as active dialog
        self._active_dialog = dialog
        # Use exec() instead of show() to make it modal and ensure it appears
        dialog.exec()
        # Clean up after dialog closes
        self.__handle_dialog_closed(dialog)
        
    def __handle_dialog_closed(self, dialog):
        """
        Common handler for when any dialog is closed or rejected.
        Clears the active dialog reference and performs cleanup.
        """
        self._active_dialog = None
        dialog.deleteLater()

    def _threadpool_setup(self):
        threadpool = QThreadPool()
        threadpool.maxThreadCount()
        return threadpool

    def _worker_setup(self):
        return Worker(self._msg_queue_handler)

    def _msg_queue_handler(self):
        notif_threads = []
        while not self.quit:
            time.sleep(0.001)
            if MessageQueue.notify_sysapp.qsize() > 0:
                msg = MessageQueue.notify_sysapp.get()
                logging.debug(f"Got a message: {msg}")
                if msg == QueueMessageType.USER_REQUEST:
                    t = threading.Thread(target=self.prompt_notification)
                    t.start()
                    notif_threads.append(t)
                    self._tray_icon.setIcon(self.auth_icon)
                    self._tray_icon.setToolTip('Requesting Authentication...')
                    # Use QMetaObject.invokeMethod to safely call a method in the main thread
                    QMetaObject.invokeMethod(self, "start_icon_reset_timer",
                                           Qt.ConnectionType.QueuedConnection)
                elif msg == QueueMessageType.AUTH_RESPONSE:
                    self.cancel_notification()
                    self._reset_icon()
                    # Stop the timer if it's running
                    if self.icon_reset_timer.isActive():
                        self.icon_reset_timer.stop()

            tempThreadList = []
            for t in notif_threads:
                if not t.is_alive():
                    t.join()
                    tempThreadList.append(t)
            for t in tempThreadList:
                notif_threads.remove(t)

    def _exit(self):
        logging.info("Sysapp Exiting")
        MessageQueue.notify_udev.put(QueueMessageType.QUIT)
        if self.notification_fw == self.NotificationFramework.NOTIFY_SEND:
            NotifySend.cancel_notification()
        self.quit = True
        self.app.quit()

    def closeEvent(self, a0):
        # Override closeEvent to hide the window instead of closing the application
        if self._tray_icon.isVisible():
            self.hide()
            if a0:
                a0.ignore()
            else: #panic!
                self._exit()
        else:
            self._exit()
            
    @pyqtSlot()
    def start_icon_reset_timer(self):
        """Start the icon reset timer from the main thread"""
        # Start the timer to reset the icon after 15 seconds (15000 ms)
        self.icon_reset_timer.start(15000)
        
    def _reset_icon(self):
        """Reset the tray icon to the main icon and update the tooltip."""
        self._tray_icon.setIcon(self.main_icon)
        self._tray_icon.setToolTip('EyeBeePasskey')

    def _finalise(self):
        self._tray_icon.setContextMenu(self.menu)
        self._tray_icon.show()
        self.threadPool.start(self.worker)
        self.launch_notification()
        self.app.exec()


class NotifySend:
    ACCEPT = 0
    DECLINE = 1
    EXPIRE = 2

    proc = None

    @classmethod
    def prompt_notification(cls):
        timeout = 15000  # Expire notification in a minute
        cmd = ['notify-send',
            '--action=accept=Accept',
            '--action=decline=Decline',
            '--action=default=default',
            '--expire-time={}'.format(timeout),
            '--icon=info',
            '--app-name=EyeBeeKey',
            'I challenge thee',
            'Pirate key recieved a webauthn ceremony, should I respond?']
        cls.proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        while cls.proc.poll() is None:
            time.sleep(0.002)
        outMsg, errMsg = cls.proc.communicate()
        outMsg, errMsg = outMsg.decode('utf-8'), errMsg.decode('utf-8')

        if outMsg == 'accept\n' or outMsg == 'default\n':
            MessageQueue.notify_auth.put(QueueMessageType.USER_RESPONSE_ACCEPT)
        else:
            MessageQueue.notify_auth.put(QueueMessageType.USER_RESPONSE_REJECT)

    @classmethod
    def cancel_notification(cls):
        if cls.proc:
            cls.proc.terminate()

    @classmethod
    def launch_notification(cls):
        cmd = ['notify-send',
            '--app-name=EyeBeeKey',
            '--icon=info', 'EyeBeeKey',
            'Starting the Pirate Passkey UHID Service']
        subprocess.Popen(cmd).communicate()

# Made with Bob



# Copyrite IBM 2022, 2025
# IBM Confidential

import datetime
import struct
import base64
import binascii

import asn1

from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.backends import default_backend
from cryptography.x509.oid import ObjectIdentifier
from cryptography import x509


class CertUtils(object):
    '''
    Class for generating certificates for FIDO2 Authenticators. methods should be treated as
    static
    '''
    TCG_KP_AIK_CERTIFICATE_ATTRIBUTE = "2.23.133.8.3"
    TPM_MANUFACTURER = "2.23.133.2.1"
    TPM_VENDOR = "2.23.133.2.2"
    TPM_FW_VERSION = "2.23.133.2.3"
    TPM_VENDOR_ID = 0xfffff1d0

    @classmethod
    def _long_to_bytes(cls, l):
        limit = 256**4 - 1  #max value we can fit into a struct.pack
        parts = []
        while l:
            parts.append(l & limit)
            l >>= 32
        parts = parts[::-1]
        return struct.pack(">" + 'L' * len(parts), *parts)

    class AAGUIDExtension(x509.UnrecognizedExtension):

        def __init__(self, aaguid, oid=ObjectIdentifier("1.3.6.1.4.1.45724.1.1.4")):
            encoder = asn1.Encoder()
            encoder.start()
            encoder.write(aaguid, nr=asn1.Numbers.OctetString)
            encodedAAGUID = encoder.output()
            super().__init__(oid, encodedAAGUID)

    class AndroidKeystoreExtension(x509.UnrecognizedExtension):
        '''
        KeyDescription ::= SEQUENCE
        {
          attestationVersion         INTEGER,
          attestationSecurityLevel   SecurityLevel,
          keymasterVersion           INTEGER,
          keymasterSecurityLevel     SecurityLevel,
          attestationChallenge       OCTET_STRING,
          uniqueId                   OCTET_STRING,
          softwareEnforced           AuthorizationList,
          teeEnforced                AuthorizationList
        }
        SecurityLevel ::= ENUMERATED
        {
          software,
          trustedenvironment,
          strongbox
        }
        AuthorizationList ::= SEQUENCE {
          purpose                     [1] EXPLICIT SET OF INTEGER OPTIONAL,
          algorithm                   [2] EXPLICIT INTEGER OPTIONAL,
          keySize                     [3] EXPLICIT INTEGER OPTIONAL,
          digest                      [5] EXPLICIT SET OF INTEGER OPTIONAL,
          padding                     [6] EXPLICIT SET OF INTEGER OPTIONAL,
          ecCurve                     [10] EXPLICIT INTEGER OPTIONAL,
          rsaPublicExponent           [200] EXPLICIT INTEGER OPTIONAL,
          rollbackResistance          [303] EXPLICIT NULL OPTIONAL,
          activeDateTime              [400] EXPLICIT INTEGER OPTIONAL,
          originationExpireDateTime   [401] EXPLICIT INTEGER OPTIONAL,
          usageExpireDateTime         [402] EXPLICIT INTEGER OPTIONAL,
          noAuthRequired              [503] EXPLICIT NULL OPTIONAL,
          userAuthType                [504] EXPLICIT INTEGER OPTIONAL,
          authTimeout                 [505] EXPLICIT INTEGER OPTIONAL,
          allowWhileOnBody            [506] EXPLICIT NULL OPTIONAL,
          trustedUserPresenceRequired [507] EXPLICIT NULL OPTIONAL,
          trustedConfirmationRequired [508] EXPLICIT NULL OPTIONAL,
          unlockedDeviceRequired      [509] EXPLICIT NULL OPTIONAL,
          allApplications             [600] EXPLICIT NULL OPTIONAL,
          applicationId               [601] EXPLICIT OCTET_STRING OPTIONAL,
          creationDateTime            [701] EXPLICIT INTEGER OPTIONAL,
          origin                      [702] EXPLICIT INTEGER OPTIONAL,
          rollbackResistant           [703] EXPLICIT NULL OPTIONAL,
          rootOfTrust                 [704] EXPLICIT RootOfTrust OPTIONAL,
          osVersion                   [705] EXPLICIT INTEGER OPTIONAL,
          osPatchLevel                [706] EXPLICIT INTEGER OPTIONAL,
          attestationApplicationId    [709] EXPLICIT OCTET_STRING OPTIONAL,
          attestationIdBrand          [710] EXPLICIT OCTET_STRING OPTIONAL,
          attestationIdDevice         [711] EXPLICIT OCTET_STRING OPTIONAL,
          attestationIdProduct        [712] EXPLICIT OCTET_STRING OPTIONAL,
          attestationIdSerial         [713] EXPLICIT OCTET_STRING OPTIONAL,
          attestationIdImei           [714] EXPLICIT OCTET_STRING OPTIONAL,
          attestationIdMeid           [715] EXPLICIT OCTET_STRING OPTIONAL,
          attestationIdManufacturer   [716] EXPLICIT OCTET_STRING OPTIONAL,
          attestationIdModel          [717] EXPLICIT OCTET_STRING OPTIONAL,
          vendorPatchLevel            [718] EXPLICIT INTEGER OPTIONAL,
          bootPatchLevel              [719] EXPLICIT INTEGER OPTIONAL
        }
        RootOfTrust ::= SEQUENCE
        {
          verifiedBootKey            OCTET_STRING,
          deviceLocked               BOOLEAN,
          verifiedBootState          VerifiedBootState,
          verifiedBootHash           OCTET_STRING
        }
        VerifiedBootState ::= ENUMERATED
        {
          verified,
          selfsigned,
          unverified,
          failed
        }


        MIICyjCCAnCgAwIBAgIBATAKBggqhkjOPQQDAjCBiDELMAkGA1UEBhMCVVMxEzARBgNVBAgMCkNhbGlm
        b3JuaWExFTATBgNVBAoMDEdvb2dsZSwgSW5jLjEQMA4GA1UECwwHQW5kcm9pZDE7MDkGA1UEAwwyQW5k
        cm9pZCBLZXlzdG9yZSBTb2Z0d2FyZSBBdHRlc3RhdGlvbiBJbnRlcm1lZGlhdGUwHhcNMTgxMjAyMDkx
        MDI1WhcNMjgxMjAyMDkxMDI1WjAfMR0wGwYDVQQDDBRBbmRyb2lkIEtleXN0b3JlIEtleTBZMBMGByqG
        SM49AgEGCCqGSM49AwEHA0IABDhJog/eJsNLAIg5GlgneD3/k4gLFlQIiq369XollUmhdDxLUkXPJoXP
        kQVDZ81Pr7lITnBZNlEBH8DcznYhxo+jggExMIIBLTALBgNVHQ8EBAMCB4AwgfwGCisGAQQB1nkCAREE
        ge0wgeoCAQIKAQACAQEKAQEEICpDgte72J2LW98Xcs/syhQ5JIe5/VcfLrcr35feBtS2BAAwgYK/gxAI
        AgYBZ24u4XC/gxEIAgYBsOqNrXC/gxIIAgYBsOqNrXC/hT0IAgYBZ24u3+i/hUVOBEwwSjEkMCIEHWNv
        bS5nb29nbGUuYXR0ZXN0YXRpb25leGFtcGxlAgEBMSIEIFrQXsIhyPg6ImEn3sVXUAw+V0vGASWp3CHL
        C+SgBmCVMDOhBTEDAgECogMCAQOjBAICAQClBTEDAgEEqgMCAQG/g3gDAgEXv4N5AwIBHr+FPgMCAQAw
        HwYDVR0jBBgwFoAUP/ys1hqxOp6BILjVJRzFZbsekakwCgYIKoZIzj0EAwIDSAAwRQIgZ3c5CJOAVf1j
        TuQT6q/CHYrHqUQb35evY5FPmzsAr/4CIQC5wMiUWMJSjisl+ojE1j3cdeG8gPuU3MYiiVLQT4EkGA==
        '''

        def __init__(self, nonce, oid=ObjectIdentifier("1.3.6.1.4.1.11129.2.1.17")):
            # https://source.android.com/security/keystore/attestation#attestation-extension
            encoder = asn1.Encoder()
            encoder.start()
            encoder.enter(asn1.Numbers.Sequence)
            encoder.write(3, nr=asn1.Numbers.Integer)  # Sequence[0] == attestationVersion
            encoder.write(0, nr=asn1.Numbers.Enumerated)  # Sequence[1] == attestationSecurityLevel
            encoder.write(1, nr=asn1.Numbers.Integer)  # Sequence[2] == keymasterVersion
            encoder.write(1, nr=asn1.Numbers.Enumerated)  # Sequence[3] == keymasterSecurityLevel
            encoder.write(nonce, nr=asn1.Numbers.OctetString)  # Sequece[4] == attestationChallenge
            encoder.write(b'9001', nr=asn1.Numbers.OctetString)  # Sequece[5] == uniqueId
            encoder.enter(asn1.Numbers.Sequence)  # Sequence[6] == softwareEnforced
            encoder.enter(400, asn1.Classes.Context)
            encoder.write(123456, nr=asn1.Numbers.Integer)
            encoder.leave()
            encoder.enter(401, asn1.Classes.Context)
            encoder.write(654321, nr=asn1.Numbers.Integer)
            encoder.leave()
            encoder.enter(701, asn1.Classes.Context)
            encoder.write(3152425, nr=asn1.Numbers.Integer)
            encoder.leave()
            encoder.leave()  # end Sequence[7]
            encoder.enter(asn1.Numbers.Sequence)  # Sequence[7] == teeEnforced
            encoder.enter(1, asn1.Classes.Context)
            encoder.enter(asn1.Numbers.Set)
            encoder.write(2, nr=asn1.Numbers.Integer)
            encoder.leave()
            encoder.leave()  #end set
            encoder.enter(3, asn1.Classes.Context)
            encoder.write(256, nr=asn1.Numbers.Integer)
            encoder.leave()
            encoder.enter(504, asn1.Classes.Context)
            encoder.write(23, nr=asn1.Numbers.Integer)
            encoder.leave()
            encoder.enter(505, asn1.Classes.Context)
            encoder.write(30, nr=asn1.Numbers.Integer)
            encoder.leave()
            encoder.enter(702, asn1.Classes.Context)
            encoder.write(0, nr=asn1.Numbers.Integer)
            encoder.leave()
            encoder.leave()  # end Sequence[7]
            encoder.leave()  # end androidKey asn1
            super().__init__(oid, encoder.output())

    class AppleNonceExtension(x509.UnrecognizedExtension):

        def __init__(self, nonce, oid=ObjectIdentifier("1.2.840.113635.100.8.2")):
            encoder = asn1.Encoder()
            encoder.start()
            encoder.enter(asn1.Numbers.Sequence)
            encoder.enter(0, asn1.Classes.Context)
            encoder.write(nonce)
            encoder.leave()
            encoder.leave()
            encodedNonce = encoder.output()
            super().__init__(oid, encodedNonce)

    @classmethod
    def __cert_builder(cls, subject=None, issuer=None, lifetime=265, serial=None, keyPair=None):
        if not issuer or not serial or not keyPair: #TPM subject is none
            raise ValueError("Missing required parameters")
        return x509.CertificateBuilder() \
                    .subject_name(subject) \
                    .issuer_name(issuer) \
                    .public_key(keyPair.get_public()) \
                    .serial_number(serial) \
                    .not_valid_before(datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(days=1)) \
                    .not_valid_after(datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(days=lifetime)) \

    @classmethod
    def __add_extensions(cls, certBuilder, extensions):
        for extension in extensions:
            certBuilder = certBuilder.add_extension(extension,
                                                    critical=True if extension.oid._name == 'subjectAltName' else False)
        return certBuilder

    @classmethod
    def get_bytes(cls, cert, encoding=serialization.Encoding.DER):
        encoded = cls.get_encoded(cert, encoding=encoding)
        #This is the ASN.1 DER Encoded certificate
        return base64.b64encode(encoded)

    @classmethod
    def load_der_certificate(cls, certBytes):
        return x509.load_der_x509_certificate(certBytes, default_backend())

    @classmethod
    def get_encoded(cls, cert, encoding=serialization.Encoding.DER):
        encoded = cert.public_bytes(encoding=encoding)
        return encoded

    @classmethod
    def gen_cert(cls,
                 subject=None,
                 issuer=None,
                 lifetime=365,
                 serial=x509.random_serial_number(),
                 extensions=None,
                 keyPair=None,
                 signKeyPair=None,
                 signer=hashes.SHA256(),
                 backend=default_backend()):
        '''
        extension should be tuple (extension, isCritical=False)
        '''
        if not keyPair:
            raise ValueError("Must provide a keypair")
        if issuer == None:  #Self signed
            issuer = subject
        if signKeyPair == None:  # self signed
            signKeyPair = keyPair

        certBuilder = cls.__cert_builder(subject, issuer, lifetime, serial, keyPair)
        certBuilder = cls.__add_extensions(certBuilder, extensions)
        return certBuilder.sign(signKeyPair.get_private(), signer, backend)

    @classmethod
    def gen_ca_cert(cls,
                    subject=None,
                    lifetime=365,
                    serial=x509.random_serial_number(),
                    keyPair=None,
                    signer=hashes.SHA256(),
                    backend=default_backend()):
        '''
        generate certificate that can be used as a ca certificate for authenticators. This
        certificate contains the ski extension
        '''
        if not keyPair:
            raise ValueError("Must provide a keypair")
        # CA cert requires basic contraint, ski, key usage and san extensions
        extensions = [
            x509.SubjectKeyIdentifier.from_public_key(keyPair.get_public()),
            x509.BasicConstraints(True, 3),
            x509.KeyUsage(True, False, False, False, False, True, True, False, False),
        ]

        return cls.gen_cert(subject, subject, lifetime, serial, extensions, keyPair, keyPair, signer, backend)

    @classmethod
    def gen_aik_cert(cls,
                     subject=None,
                     issuer=None,
                     lifetime=365,
                     serial=x509.random_serial_number(),
                     keyPair=None,
                     signKeyPair=None,
                     aaguid=None,
                     san=None,
                     androidKeyNonce=None,
                     signer=hashes.SHA256(),
                     backend=default_backend()):
        '''
        Generate Leaf cert in trust chain
        issuer should match the keyPair used to sign the certificate
        '''
        if san is None:
            sanId = cls._long_to_bytes(cls.TPM_VENDOR_ID)
            san = x509.Name([
                x509.NameAttribute(ObjectIdentifier(cls.TPM_MANUFACTURER), u"IBM"),
                x509.NameAttribute(ObjectIdentifier(cls.TPM_VENDOR), u"id:{}".format(binascii.b2a_uu(sanId))),
                x509.NameAttribute(ObjectIdentifier(cls.TPM_FW_VERSION), u"id:1")
            ])
        extensions = [
            x509.BasicConstraints(False, None),
            x509.KeyUsage(True, True, False, True, False, True, True, False, False),
            x509.ExtendedKeyUsage([ObjectIdentifier(cls.TCG_KP_AIK_CERTIFICATE_ATTRIBUTE)]),
            x509.SubjectAlternativeName([x509.DirectoryName(san)])
        ]
        if aaguid is not None:
            extensions += [CertUtils.AAGUIDExtension(aaguid)]
        if androidKeyNonce is not None:
            extensions += [CertUtils.AndroidKeystoreExtension(androidKeyNonce)]

        return cls.gen_cert(subject, issuer, lifetime, serial, extensions, keyPair, signKeyPair, signer, backend)

    @classmethod
    def gen_intermedaite_cert(cls,
                              subject=None,
                              issuer=None,
                              lifetime=365,
                              serial=x509.random_serial_number(),
                              keyPair=None,
                              signKeyPair=None,
                              signer=hashes.SHA256(),
                              backend=default_backend()):
        '''
        Generate intermediate certificate in trust chain
        '''
        extensions = [
            x509.BasicConstraints(True, 3),
            x509.KeyUsage(True, True, False, True, False, True, True, False, False),
            x509.ExtendedKeyUsage([ObjectIdentifier(cls.TCG_KP_AIK_CERTIFICATE_ATTRIBUTE)])
        ]

        return cls.gen_cert(subject, issuer, lifetime, serial, extensions, keyPair, signKeyPair, signer, backend)

    @classmethod
    def gen_apple_cert(cls,
                       subject=None,
                       issuer=None,
                       lifetime=365,
                       serial=x509.random_serial_number(),
                       keyPair=None,
                       signKeyPair=None,
                       nonce=None,
                       signer=hashes.SHA256(),
                       backend=default_backend()):
        '''
        Generate Apple Attestation certificate. At the moment this is just a x509 with some apple extension
        which I am sure is very useful to apple.
        '''
        extensions = [CertUtils.AppleNonceExtension(nonce)]
        return cls.gen_cert(subject, issuer, lifetime, serial, extensions, keyPair, signKeyPair, signer, backend)



"""
Compatibility layer for decrypting credential IDs encrypted with either Fernet or SymmetricKey.

This module provides a class that can detect and decrypt credential IDs that were encrypted
with either the legacy Fernet encryption or the new SymmetricKey encryption.
"""

import base64
import logging
from typing import Optional, Union

from cryptography.fernet import Fernet
from soft_fido2.symmetric_key import SymmetricKey
from soft_fido2.key_pair import KeyPair

class CredentialIdMigrator:
    """
    A class that can decrypt credential IDs encrypted with either Fernet or SymmetricKey.
    
    This class provides backward compatibility for existing Fernet-encrypted credential IDs
    while supporting the new SymmetricKey encryption for new registrations.

    This prvodes a path forwards for legacy registrations which use a Fernet key.
    """
    
    @staticmethod
    def _urlb64_decode(b64_string: Union[str, bytes]) -> bytes:
        """
        Decode URL-safe base64 strings which may be missing trailing padding.

        Args:
            b64_string: URL-safe base64 encoded string to decode

        Returns:
            Decoded bytes
        """
        if isinstance(b64_string, str):
            b64_string = b64_string.encode('utf-8')
            
        # Add padding if needed
        pad = len(b64_string) % 4
        if pad:
            b64_string += b'=' * pad
            
        return base64.urlsafe_b64decode(b64_string)
    
    @classmethod
    def decrypt_credential_id(cls, cred_id: Union[str, bytes], seed: bytes) -> Optional[KeyPair]:
        """
        Try to decrypt a credential ID using either a Fernet or SymmetricKey.
        
        Args:
            cred_id: URL-safe base64 encoded credential ID
            seed: The seed used to encrypt the credential ID
            
        Returns:
            The decrypted KeyPair or None if decryption fails
        """
        if isinstance(cred_id, str): 
            cred_id = cls._urlb64_decode(cred_id)
        fKey = Fernet(base64.urlsafe_b64encode(seed))
        sKey = SymmetricKey(base64.urlsafe_b64encode(seed).decode())

        if len(cred_id) > 0:
            # Try SymmetricKey first
            try:
                key_bytes = sKey.decrypt(cred_id)
                return KeyPair.load_key_pair(key_bytes)
            except Exception as e:
                logging.warning(f"Failed to decrypt with SymmetricKey: {e}")
            
            # Try Fernet if SymmetricKey failed
            try:
                key_bytes = fKey.decrypt(cred_id)
                return KeyPair.load_key_pair(key_bytes)
            except Exception as e:
                logging.warning(f"Failed to decrypt with Fernet: {e}")
        
        return None
    
    @classmethod
    def migrate_credential_id(cls, cred_id: str, seed: bytes) -> Optional[str]:
        """
        Migrate a credential ID from Fernet encryption to SymmetricKey encryption.
        
        Args:
            cred_id: URL-safe base64 encoded credential ID
            seed: The seed used to generate the Fernet and Symmetric keys
            
        Returns:
            The re-encrypted credential ID or None if migration fails
        """
        # First decrypt the credential ID using the old key
        key_pair = cls.decrypt_credential_id(cred_id, seed)
        
        if key_pair is None:
            return None
        
        # Re-encrypt the private key using the new SymmetricKey
        new_cred_id_bytes = SymmetricKey(seed).encrypt(key_pair.get_private_bytes())
        
        # Return the base64-encoded credential ID
        return base64.urlsafe_b64encode(new_cred_id_bytes).decode('utf-8')

# Made with Bob



"""
Custom implementation of symmetric encryption with AES-GCM.
This replaces the dependency on cryptography.fernet with a custom implementation.

This implementation provides authenticated encryption that guarantees a message
cannot be manipulated or read without the key.
"""

import base64
import os
import struct
import time

from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends import default_backend

class SymmetricKey:
    """
    SymmetricKey implements symmetric encryption using AES-GCM.
    
    This class provides methods to generate keys, encrypt, and decrypt messages
    using AES-GCM for authenticated encryption.
    """
    
    # Version byte for our GCM-based tokens (different from standard Fernet)
    VERSION = b'\x81'  # Using 0x81 to distinguish from standard Fernet's 0x80
    
    # Minimum token size: Version (1) + Timestamp (8) + Nonce (12) + Tag (16) + min ciphertext (1)
    MIN_TOKEN_SIZE = 38
    
    def __init__(self, key):
        """
        Initialize with a symmetric key.
        
        Args:
            key (bytes or str): A 32-byte key or a URL-safe base64-encoded 32-byte key.
                Used for both encryption and authentication in GCM mode.
        """
        if isinstance(key, str):
            key = base64.urlsafe_b64decode(key.encode('utf-8'))
        
        if len(key) != 32:
            raise ValueError(
                "Symmetric key must be 32 bytes (URL-safe base64-encoded)."
            )
        
        self._key = key
        self._backend = default_backend()
    
    @classmethod
    def generate_key(cls):
        """
        Generates a new symmetric key.
        
        Returns:
            str: A URL-safe base64-encoded 32-byte key.
        """
        key = os.urandom(32)
        return base64.urlsafe_b64encode(key).decode('utf-8')
    
    def encrypt(self, data):
        """
        Encrypts data using AES-GCM.
        
        Args:
            data (bytes): The data to encrypt.
            
        Returns:
            bytes: The encrypted token.
        """
        if not isinstance(data, bytes):
            raise TypeError("data must be bytes.")
        
        # Current time in seconds since the epoch, as a 64-bit integer
        current_time = int(time.time())
        
        # Generate a random 96-bit nonce (recommended size for GCM)
        nonce = os.urandom(12)
        
        # Encrypt the data with AES-GCM
        encryptor = Cipher(
            algorithms.AES(self._key),
            modes.GCM(nonce),
            backend=self._backend
        ).encryptor()
        
        # Include timestamp in the associated data for authentication
        associated_data = struct.pack('>Q', current_time)
        encryptor.authenticate_additional_data(associated_data)
        ciphertext = encryptor.update(data) + encryptor.finalize()
        tag = encryptor.tag
        
        # Format: Version (0x81) || Timestamp (8 bytes) || Nonce (12 bytes) || Tag (16 bytes) || Ciphertext
        token = (
            self.VERSION +
            associated_data +  # Timestamp
            nonce +
            tag +
            ciphertext
        )
        
        # Return URL-safe base64 encoded token
        return base64.urlsafe_b64encode(token)
    
    def _parse_token(self, token_bytes):
        """
        Parse a token into its components.
        
        Args:
            token_bytes: Raw bytes of the token after base64 decoding
            
        Returns:
            tuple: (version, timestamp_bytes, nonce, tag, ciphertext)
            
        Raises:
            ValueError: If the token format is invalid
        """
        if len(token_bytes) < self.MIN_TOKEN_SIZE:
            raise ValueError(f"Token too short: {len(token_bytes)} bytes, minimum is {self.MIN_TOKEN_SIZE}")
        
        # Extract parts
        version = token_bytes[0:1]
        timestamp_bytes = token_bytes[1:9]
        nonce = token_bytes[9:21]
        tag = token_bytes[21:37]
        ciphertext = token_bytes[37:]
        
        # Verify version
        if version != self.VERSION:
            raise ValueError(f"Invalid token version: {version!r}, expected {self.VERSION!r}")
        
        return version, timestamp_bytes, nonce, tag, ciphertext
    
    def _verify_timestamp(self, timestamp_bytes, ttl):
        """
        Verify that a token's timestamp is within the TTL.
        
        Args:
            timestamp_bytes: 8-byte timestamp from the token
            ttl: Time-to-live in seconds
            
        Raises:
            ValueError: If the token has expired
        """
        if ttl is None:
            return
            
        timestamp = struct.unpack('>Q', timestamp_bytes)[0]
        current_time = int(time.time())
        
        if current_time - timestamp > ttl:
            age = current_time - timestamp
            raise ValueError(f"Token expired: token is {age} seconds old, but TTL is {ttl}")
    
    def decrypt(self, token, ttl=None):
        """
        Decrypts a token.
        
        Args:
            token (bytes or str): The token to decrypt.
            ttl (int, optional): Time-to-live in seconds. If the token is older
                than this, decryption will fail. If None, no TTL check is performed.
                    
        Returns:
            bytes: The decrypted data.
                
        Raises:
            ValueError: If the token is invalid, expired, or cannot be decrypted
            TypeError: If the token is not bytes or str
        """
        if isinstance(token, str):
            token = token.encode('utf-8')
        
        try:
            data = base64.urlsafe_b64decode(token)
        except (TypeError, ValueError) as e:
            raise ValueError(f"Invalid token: base64 decoding failed: {e}")
        
        # Parse the token into its components
        _, timestamp_bytes, nonce, tag, ciphertext = self._parse_token(data)
        
        # Verify timestamp if TTL is provided
        self._verify_timestamp(timestamp_bytes, ttl)
        
        # Decrypt the ciphertext with AES-GCM
        try:
            decryptor = Cipher(
                algorithms.AES(self._key),
                modes.GCM(nonce, tag),
                backend=self._backend
            ).decryptor()
            
            # Include timestamp in the associated data for authentication
            decryptor.authenticate_additional_data(timestamp_bytes)
            
            # Decrypt and verify the ciphertext
            plaintext = decryptor.update(ciphertext) + decryptor.finalize()
            
            return plaintext
        except Exception as e:
            # Catch any other exceptions, including cryptography.exceptions.InvalidTag
            # This is important for the test_invalid_token test case
            raise ValueError(f"Decryption failed: {str(e)}")

# Made with Bob



# Copyrite IBM 2022, 2025
# IBM Confidential

import cbor2 as cbor


class CaBLE(object):

    def __init__(self):
        return


    def decode_qr_data(self, data):
        chunkSize = 17
        chunkBytes = 7
        buff = b''
        for i in range( int(len(data)/chunkSize) ):
            buff += (int(data[i * chunkSize: (i + 1) * chunkSize])).to_bytes(chunkBytes, 'little')
        if len(data) % chunkSize != 0:
            #unpack the trailing bytes
            buff += (int(data[int(len(data)/chunkSize):])).to_bytes(chunkBytes, 'little')
        return cbor.loads(buff)



# Copyrite IBM 2022, 2025
# IBM Confidential

import queue


class QueueMessageType:
    USER_REQUEST = 0
    USER_RESPONSE_ACCEPT = 1
    USER_RESPONSE_REJECT = 2
    USER_RESPONSE_TIMEOUT = 3
    AUTH_REQUEST = 4
    AUTH_RESPONSE = 5
    KEEPALIVE = 6
    KEEPALIVE_CANCEL = 7
    QUIT = 8
    CLOSE_EVENT = 9


class MessageQueue:
    ''' read by uhid_device.py '''
    notify_udev = queue.Queue(maxsize=100)
    ''' read by systray_app.py '''
    notify_sysapp = queue.Queue(maxsize=100)
    ''' read by passkey_device.py.Authenticator '''
    notify_auth = queue.Queue(maxsize=100)